import uuid

from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.core.serializers.json import DjangoJSONEncoder
from django.db import models
from django.utils.translation import gettext
from django.utils.translation import gettext_lazy as _
from wbcore.contrib.authentication.models import User
from wbcore.contrib.color.enums import WBColor
from wbcore.models import WBModel


class Process(WBModel):
    """Starting a workflow creates a process"""

    class ProcessState(models.TextChoices):
        ACTIVE = "Active", _("Active")
        FINISHED = "Finished", _("Finished")
        FAILED = "Failed", _("Failed")

        @classmethod
        def get_color_map(cls):
            colors = [
                WBColor.BLUE_LIGHT.value,
                WBColor.GREEN_LIGHT.value,
                WBColor.RED_LIGHT.value,
            ]
            return [choice for choice in zip(cls, colors)]

    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False, verbose_name=_("UUID"))
    workflow = models.ForeignKey(
        to="workflow.Workflow", on_delete=models.PROTECT, verbose_name=_("Workflow"), editable=False
    )
    started = models.DateTimeField(auto_now_add=True, verbose_name=_("Started"))
    finished = models.DateTimeField(verbose_name=_("Finished"), null=True, blank=True, editable=False)
    instance_id = models.CharField(blank=True, null=True, editable=False, verbose_name=_("Instance"), max_length=128)
    content_type = models.ForeignKey(
        to=ContentType,
        on_delete=models.SET_NULL,
        null=True,
        blank=True,
        verbose_name=_("Attached Model"),
        editable=False,
    )
    instance = GenericForeignKey("content_type", "instance_id")
    preserved_instance = models.JSONField(
        verbose_name=_("Preserved Instance"), null=True, blank=True, encoder=DjangoJSONEncoder
    )
    state = models.CharField(
        max_length=64,
        choices=ProcessState.choices,
        verbose_name=_("State"),
        help_text=_("The current state of this process. Can be one of: 'Active', 'Finished' or 'Failed'"),
        editable=False,
        default=ProcessState.ACTIVE,
    )

    def set_failed(self):
        """Sets the entire process and all remaining active process steps to failed"""
        from wbcore.contrib.workflow.models.step import Step, UserStep

        # Set all active process steps to failed
        for active_process_step in self.process_steps.filter(
            state__in=[ProcessStep.StepState.ACTIVE, ProcessStep.StepState.WAITING]
        ):
            # We cannot use "set_failed" or we run into infinite loops
            active_process_step.state = ProcessStep.StepState.FAILED
            error_message = gettext("Entire process failed due to unreachable steps")
            active_process_step.error_message = error_message
            active_process_step.save()

            # Notify assignees if user step
            if active_process_step.step.step_type == Step.StepType.USERSTEP:
                step: UserStep = active_process_step.step.get_casted_step()
                if step.notify_user and (all_assignees := active_process_step.get_all_assignees()):
                    step.notify_assignees(all_assignees, active_process_step, error_message)

        # Set process to failed
        self.state = self.ProcessState.FAILED
        self.save()

    def __str__(self) -> str:
        return _("Process {} for workflow {}").format(str(self.id), self.workflow.name)

    @classmethod
    def get_endpoint_basename(cls):
        return "wbcore:workflow:process"

    @classmethod
    def get_representation_value_key(cls):
        return "id"

    @classmethod
    def get_representation_endpoint(cls):
        return "wbcore:workflow:processrepresentation-list"

    @classmethod
    def get_representation_label_key(cls):
        return "{{id}}"

    class Meta:
        verbose_name = _("Process")
        verbose_name_plural = _("Processes")


class ProcessStep(WBModel):
    """The running instance of a step created after a process was started"""

    class StepState(models.TextChoices):
        ACTIVE = "Active", _("Active")
        FINISHED = "Finished", _("Finished")
        WAITING = "Waiting", _("Waiting")
        FAILED = "Failed", _("Failed")
        CANCELED = "Canceled", _("Canceled")

        @classmethod
        def get_color_map(cls):
            colors = [
                WBColor.BLUE_LIGHT.value,
                WBColor.GREEN_LIGHT.value,
                WBColor.YELLOW_LIGHT.value,
                WBColor.RED_LIGHT.value,
                WBColor.GREY.value,
            ]
            return [choice for choice in zip(cls, colors)]

    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False, verbose_name=_("UUID"))
    process = models.ForeignKey(
        to=Process, on_delete=models.CASCADE, verbose_name=_("Process"), editable=False, related_name="process_steps"
    )
    step = models.ForeignKey(
        to="workflow.Step",
        on_delete=models.CASCADE,
        verbose_name=_("Step"),
        editable=False,
        related_name="process_steps",
    )
    started = models.DateTimeField(auto_now_add=True, verbose_name=_("Started"))
    finished = models.DateTimeField(verbose_name=_("Finished"), null=True, blank=True, editable=False)
    assignee = models.ForeignKey(
        to="authentication.User",
        verbose_name=_("Assignee"),
        on_delete=models.SET_NULL,
        related_name="assigned_process_steps",
        editable=False,
        null=True,
        blank=True,
    )
    group = models.ForeignKey(
        to="authentication.Group",
        verbose_name=_("Group"),
        on_delete=models.SET_NULL,
        related_name="assigned_process_steps",
        editable=False,
        null=True,
        blank=True,
    )
    permission = models.ForeignKey(
        to="authentication.Permission",
        verbose_name=_("Permission"),
        on_delete=models.SET_NULL,
        blank=True,
        null=True,
        editable=False,
        related_name="related_process_steps",
        help_text=_("The permission needed to be able to view this step being executed."),
    )
    state = models.CharField(
        max_length=64,
        choices=StepState.choices,
        verbose_name=_("State"),
        help_text=_(
            "The current state of this process step. Can be one of: 'Active', 'Finished', 'Waiting', 'Failed' or 'Canceled'"
        ),
        editable=False,
        default=StepState.ACTIVE,
    )
    status = models.CharField(
        max_length=64,
        verbose_name=_("Status"),
        help_text=_("The status that will be set in the attached model upon transitioning to this step"),
        null=True,
        blank=True,
        editable=False,
    )
    error_message = models.CharField(
        max_length=128,
        verbose_name=_("Error Message"),
        help_text=_("An error message will be displayed when this step fails"),
        null=True,
        blank=True,
        editable=False,
    )

    def get_instance(self):
        """Returns the attached instance that shall be checked by a condition

        Returns:
            The attached instance
        """

        if self.process.instance:
            return self.process.instance
        # elif self.process.data_values.exists():
        # TODO: How do we integrate data values into this logic?
        return None

    def get_all_assignees(self) -> list[User]:
        """Returns the selected assignee or every member of the selected group of this process step

        Returns:
            list[User] | None: A list of assigned users
        """

        if assignee := self.assignee:
            return [assignee]
        elif group := self.group:
            return list(group.user_set.all())
        return []

    def __str__(self) -> str:
        return _("Process step {} for step {}").format(str(self.id), self.step.name)

    @classmethod
    def get_endpoint_basename(cls):
        return "wbcore:workflow:processstep"

    @classmethod
    def get_representation_value_key(cls):
        return "id"

    @classmethod
    def get_representation_endpoint(cls):
        return "wbcore:workflow:processsteprepresentation-list"

    @classmethod
    def get_representation_label_key(cls):
        return "{{id}}"

    class Meta:
        verbose_name = _("Process Step")
        verbose_name_plural = _("Process Steps")



import factory
from wbcore.contrib.notifications.factories.notification_types import (
    NotificationTypeModelFactory,
)
from wbcore.contrib.notifications.models import Notification


class NotificationModelFactory(factory.django.DjangoModelFactory):
    title = factory.Faker("pystr")
    body = factory.Faker("pystr")
    notification_type = factory.SubFactory(NotificationTypeModelFactory)
    user = factory.SubFactory("wbcore.contrib.authentication.factories.UserFactory")
    endpoint = None

    class Meta:
        model = Notification



from django_filters import Filter
from wbcore.filters.mixins import WBCoreFilterMixin


class MultipleLookupFilter(WBCoreFilterMixin):
    def __init__(self, field_class, lookup_expr, **kwargs):
        self.field_class = field_class
        self.lookup_expr = lookup_expr
        self.kwargs = kwargs

    def get_filters(self, field_name) -> dict[str, Filter]:
        filters = dict()
        for lookup_expr in self.lookup_expr:
            filters[f"{field_name}__{lookup_expr}"] = self.field_class(
                lookup_expr=lookup_expr,
                **self.kwargs,
            )

        return filters



from datetime import date, datetime
from unittest.mock import patch

import pytest
from wbcore.contrib.workflow.models import Data


@pytest.mark.django_db
class TestData:
    @pytest.mark.parametrize(
        ("type", "field_name"),
        [
            (Data.DataType.CHAR, "CharField"),
            (Data.DataType.BOOL, "BooleanField"),
            (Data.DataType.DATE, "DateField"),
            (Data.DataType.DATETIME, "DateTimeField"),
            (Data.DataType.INT, "IntegerField"),
        ],
    )
    @patch("wbcore.contrib.workflow.models.data.Data.cast_value_to_datatype")
    def test_get_serializer_field(self, mock_cast, data_factory, type, field_name):
        data = data_factory(help_text=None, data_type=type)
        field = data.get_serializer_field()
        assert field.__class__.__name__ == field_name
        assert field.label == data.label
        assert field.required == data.required
        assert "default" not in field._kwargs
        assert "help_text" not in field._kwargs
        assert not mock_cast.called

    @patch("wbcore.contrib.workflow.models.data.Data.cast_value_to_datatype")
    def test_get_serializer_field_help_text(self, mock_cast, data_factory):
        data = data_factory()
        field = data.get_serializer_field()
        assert field.label == data.label
        assert field.required == data.required
        assert "default" not in field._kwargs
        assert field.help_text == data.help_text
        assert not mock_cast.called

    @patch("wbcore.contrib.workflow.models.data.Data.cast_value_to_datatype")
    def test_get_serializer_field_default(self, mock_cast, data_factory):
        data_type = Data.DataType.INT
        default_value = "50"
        casted_default_value = 50
        data = data_factory(default=default_value, data_type=data_type)
        mock_cast.return_value = casted_default_value
        field = data.get_serializer_field()
        assert mock_cast.call_args.args == (data_type, default_value)
        assert field.label == data.label
        assert field.required == data.required
        assert field.default == casted_default_value
        assert field.help_text == data.help_text

    @pytest.mark.parametrize(
        ("data_type", "value", "expected"),
        [
            (Data.DataType.BOOL, "True", True),
            (Data.DataType.BOOL, "true", True),
            (Data.DataType.BOOL, "False", False),
            (Data.DataType.BOOL, "false", False),
            (Data.DataType.DATE, "12.07.1054", date(1054, 7, 12)),
            (Data.DataType.DATETIME, "12.07.1054 13:47:09", datetime(1054, 7, 12, 13, 47, 9)),
            (Data.DataType.CHAR, "123456", "123456"),
            (Data.DataType.INT, "123456", 123456),
        ],
    )
    def test_cast_value_to_datatype(self, data_type, value, expected):
        assert Data.cast_value_to_datatype(data_type, value) == expected

    @pytest.mark.parametrize(
        ("data_type", "value"),
        [
            (Data.DataType.BOOL, "Test"),
            (Data.DataType.BOOL, "1"),
            (Data.DataType.BOOL, "0"),
            (Data.DataType.DATE, "07/12/1054"),
            (Data.DataType.DATETIME, "07/12/1054 13:47:09"),
            (Data.DataType.INT, "Test"),
        ],
    )
    def test_cast_value_to_datatype_error(self, data_type, value):
        with pytest.raises(ValueError):
            Data.cast_value_to_datatype(data_type, value)

    @pytest.mark.parametrize(
        ("target", "data_type"),
        [
            (datetime(2003, 8, 28, 7, 18, 4), Data.DataType.DATETIME),
            (date(2003, 8, 28), Data.DataType.DATE),
            (True, Data.DataType.BOOL),
            (False, Data.DataType.BOOL),
            (69, Data.DataType.INT),
            ("Test", Data.DataType.CHAR),
        ],
    )
    @patch("wbcore.contrib.workflow.models.data.Data.cast_value_to_datatype")
    def test_cast_value_from_target_object(self, mock_cast, target, data_type):
        Data.cast_value_from_target_object(target, "Test")
        assert mock_cast.call_args.args == (data_type, "Test")

    @pytest.mark.parametrize(
        "target",
        [None, Data, str, {}, []],
    )
    def test_cast_value_from_target_object_error(self, target):
        with pytest.raises(ValueError):
            Data.cast_value_from_target_object(target, "Test")

    @pytest.mark.parametrize(
        ("data_object", "data_type", "expected"),
        [
            (datetime(2003, 8, 28, 7, 18, 4), Data.DataType.DATETIME, "28.08.2003 07:18:04"),
            (date(2003, 8, 28), Data.DataType.DATE, "28.08.2003"),
            (True, Data.DataType.BOOL, "True"),
            (False, Data.DataType.BOOL, "False"),
            (69, Data.DataType.INT, "69"),
            ("Test", Data.DataType.CHAR, "Test"),
        ],
    )
    def test_cast_datatype_to_string(self, data_type, data_object, expected):
        assert Data.cast_datatype_to_string(data_type, data_object) == expected

    @pytest.mark.parametrize(
        ("data_object", "data_type"),
        [
            (69, Data.DataType.DATETIME),
            ("Test", Data.DataType.DATE),
            (datetime(2003, 8, 28, 7, 18, 4), Data.DataType.BOOL),
        ],
    )
    def test_cast_datatype_to_string_error(self, data_type, data_object):
        with pytest.raises(ValueError):
            Data.cast_datatype_to_string(data_type, data_object)



from typing import TYPE_CHECKING, Generator

if TYPE_CHECKING:
    from django.db.models import QuerySet
    from wbcore.contrib.dataloader.models.entities import Entity


class Dataloader:
    """Base class for dataloaders, providing a foundational structure for subclasses.

    This class includes an attribute `entities`, which is a Django queryset of Entity objects.
    It is designed to be subclassed by other classes which will implement specific methods.

    Attributes:
        entities (QuerySet[Entity]): A Django queryset of Entity objects.
    """

    def __init__(self, entities: "QuerySet[Entity]"):
        self.entities = entities
        super().__init__()

    @property
    def entity_ids(self) -> Generator[int, None, None]:
        return self.entities.values_list("id", flat=True)



from unittest.mock import patch

import pytest
from wbcore.contrib.example_app.models import Match
from wbcore.contrib.example_app.viewsets.displays.match import (
    get_match_status_formatting,
    get_match_status_legend,
)


@pytest.mark.django_db
class TestMatch:
    @patch("wbcore.contrib.example_app.models.Match.MatchStatus.get_color_map")
    def test_get_match_status_legend(self, mock_color_map):
        mock_color_map.return_value = [
            (Match.MatchStatus.FINISHED, "firstcolor"),
            (Match.MatchStatus.ONGOING, "secondcolor"),
            (Match.MatchStatus.SCHEDULED, "lastcolor"),
        ]
        assert [(y.icon, y.label, y.value) for y in get_match_status_legend()[0].items] == [
            ("firstcolor", Match.MatchStatus.FINISHED.label, Match.MatchStatus.FINISHED.value),
            ("secondcolor", Match.MatchStatus.ONGOING.label, Match.MatchStatus.ONGOING.value),
            ("lastcolor", Match.MatchStatus.SCHEDULED.label, Match.MatchStatus.SCHEDULED.value),
        ]
        assert mock_color_map.called

    @patch("wbcore.contrib.example_app.models.Match.MatchStatus.get_color_map")
    def test_get_match_status_formatting(self, mock_color_map):
        mock_color_map.return_value = [
            (Match.MatchStatus.ONGOING, "firstcolor"),
            (Match.MatchStatus.SCHEDULED, "secondcolor"),
            (Match.MatchStatus.FINISHED, "lastcolor"),
        ]
        assert [
            (y.style["backgroundColor"], y.condition[1]) for y in get_match_status_formatting()[0].formatting_rules
        ] == [
            ("firstcolor", Match.MatchStatus.ONGOING.value),
            ("secondcolor", Match.MatchStatus.SCHEDULED.value),
            ("lastcolor", Match.MatchStatus.FINISHED.value),
        ]



from django.utils.translation import gettext as _
from dynamic_preferences.preferences import Section
from dynamic_preferences.registries import global_preferences_registry
from dynamic_preferences.types import IntegerPreference, StringPreference

wbcore = Section("wbcore")


@global_preferences_registry.register
class RetentionPeriod(IntegerPreference):
    section = wbcore
    name = "retention_period"
    default = 365

    verbose_name = _("Retention Period in Days")
    help_text = _(
        "When an object cannot be deleted and is disabled instead, it gets hidden from the queryset but not deleted. For compliance reasons we enable the retention for a specific period of days (defaults to a year)"
    )


@global_preferences_registry.register
class SystemUserEmailPeriod(StringPreference):
    section = wbcore
    name = "system_user_email"

    default = "system@stainly-bench.com"

    verbose_name = _("System User Email")
    help_text = _("System User Email")



from django.utils.translation import gettext as _
from wbcore.metadata.configs import display as dp
from wbcore.metadata.configs.display import Display
from wbcore.metadata.configs.display.instance_display.shortcuts import (
    create_simple_display,
)
from wbcore.metadata.configs.display.list_display import ListDisplay
from wbcore.metadata.configs.display.view_config import DisplayViewConfig


class RoleDisplayConfig(DisplayViewConfig):
    def get_list_display(self) -> ListDisplay:
        return dp.ListDisplay(fields=(dp.Field(key="title", label=_("Title")),))

    def get_instance_display(self) -> Display:
        return create_simple_display([["title"]])



import itertools
from urllib.parse import quote

import django_filters
from django_filters.filters import MultipleChoiceField
from wbcore.filters.mixins import WBCoreFilterMixin


class WBCoreChoiceFilterMixin(WBCoreFilterMixin):
    def __init__(self, *args, **kwargs):
        self.choices = kwargs["choices"]
        super().__init__(*args, **kwargs)

    def get_representation(self, request, name, view):
        representation, lookup_expr = super().get_representation(request, name, view)
        lookup_expr["input_properties"]["choices"] = list()
        if callable(self.choices):
            choice_keys = [
                choice[0] for choice in self.field.choices
            ]  # we save the choices key for validation in case we narrow down the choices list in get_representation
            # we get a new choices set but make sure its keys are still valid regarding the initial choices
            choices = list(filter(lambda x: x[0] in choice_keys, self.choices(request, view)))
        else:
            choices = self.choices

        for choice in choices:
            lookup_expr["input_properties"]["choices"].append(
                {"value": quote(choice[0]) if isinstance(choice[0], str) else choice[0], "label": choice[1]}
            )
        return representation, lookup_expr


class ChoiceFilter(WBCoreChoiceFilterMixin, django_filters.ChoiceFilter):
    filter_type = "select"


class WBCoreMultipleChoiceField(MultipleChoiceField):
    def to_python(self, value):
        # We do this because DRF wrap a coma seperated list of choices into a one list element
        if value and isinstance(value, list):
            value = list(itertools.chain(*[v.split(",") for v in value]))
        return super().to_python(value)


class MultipleChoiceFilter(WBCoreChoiceFilterMixin, django_filters.MultipleChoiceFilter):
    field_class = WBCoreMultipleChoiceField
    filter_type = "select"

    def _validate_default_with_request(self, default, request, name):
        if request_default := request.GET.get(name):
            return request_default.split(",")
        return default

    def get_representation(self, request, name, view):
        representation, lookup_expr = super().get_representation(request, name, view)
        lookup_expr["input_properties"]["multiple"] = True
        if (default := lookup_expr["input_properties"].get("default")) and not isinstance(default, list):
            lookup_expr["input_properties"]["default"] = [default]
        if (initial := lookup_expr["input_properties"].get("initial")) and not isinstance(initial, list):
            lookup_expr["input_properties"]["initial"] = [initial]
        return representation, lookup_expr



import importlib
import sys
from contextlib import suppress

from django.conf import settings
from django.utils.module_loading import import_string
from rest_framework.permissions import IsAuthenticated
from rest_framework.request import Request
from rest_framework.response import Response
from rest_framework.views import APIView

from .registry import default_registry


class MenuAPIView(APIView):
    @property
    def permission_classes(self):
        wbcore_auth = import_string(settings.WBCORE_DEFAULT_AUTH_CONFIG)(None)

        if wbcore_auth["type"] is None:
            return []

        return [IsAuthenticated]

    def get(self, request: Request) -> Response:
        # Default menu is imported from menu. In case we need multiple menus (or provide a temporary new menu) we can
        # steer this through an optional GET parameter

        mod_name = f"{settings.PROJECT_NAME}.{request.COOKIES.get('menu', 'menu')}"

        default_registry.clear()

        # Trying to import the menu from the project folder (only if it is called config)
        with suppress(ModuleNotFoundError):
            if mod := sys.modules.get(mod_name, None):
                importlib.reload(mod)
            else:
                importlib.import_module(mod_name)

        default_registry.request = request
        return Response(list(default_registry))



# Generated by Django 4.1.8 on 2023-04-12 12:56

from django.db import migrations, models


class Migration(migrations.Migration):
    initial = True

    dependencies = [
        ("directory", "0001_initial"),
        ("agenda", "0001_initial"),
    ]

    operations = [
        migrations.AddField(
            model_name="calendaritem",
            name="entities",
            field=models.ManyToManyField(
                related_name="calendar_entities", to="directory.entry", verbose_name="Entities"
            ),
        ),
        migrations.AddIndex(
            model_name="calendaritem",
            index=models.Index(fields=["period"], name="agenda_cale_period_e7e49e_idx"),
        ),
    ]



from django.conf import settings
from django.contrib.auth import get_user_model
from django.db import transaction
from django.dispatch import receiver
from django.utils import timezone
from django.utils.module_loading import import_string
from django.utils.translation import gettext
from rest_framework.reverse import reverse
from wbcore.shares.signals import handle_widget_sharing

from .models import Notification, NotificationType, NotificationTypeSetting
from .tasks import send_notification_task

User = get_user_model()


def send_notification(
    code: str,
    title: str,
    body: str,
    user: User,
    reverse_name: str | None = None,
    reverse_args=None,
    reverse_kwargs=None,
    endpoint: str | None = None,
):
    """Method for dispatching a notification

    Args:
        code: The code pointing to a `NotificationType`
        title: The title of the notification
        body: The text of the notification
        user: The user that should receive the notification
        reverse_name: The reverse name of an endpoint attached to this notification
        reverse_args: The arguments passed to the `reverse` function
        reverse_kwargs: The keyword arguments passed to the `reverse` function
        endpoint: The endpoint of resource. If provided, reverse_name + args + kwargs are disregarded
    """
    notification_user_settings = NotificationTypeSetting.objects.filter(
        notification_type__code=code,
        user=user,
    )
    if user.is_active and notification_user_settings.exists():
        if not endpoint:
            endpoint = reverse(reverse_name, reverse_args, reverse_kwargs) if reverse_name else None
        notification = Notification.objects.create(
            title=title,
            body=body,
            user=user,
            notification_type=NotificationType.objects.get(code=code),
            endpoint=endpoint,
            sent=timezone.now(),
        )
        transaction.on_commit(lambda: send_notification_task.delay(notification.pk))


@receiver(handle_widget_sharing)
def share_notification(
    request, widget_relative_endpoint, share=False, share_message=None, share_recipients=None, **kwargs
):
    if share and share_recipients and share_message:
        for recipient in share_recipients:
            send_notification(
                code="workbench.system",
                title=gettext("{} shared a widget with you").format(
                    import_string(settings.WBCORE_DEFAULT_USER_NAME)(request.user)
                ),
                user=recipient,
                body=share_message,
                endpoint=widget_relative_endpoint,
            )



from wbcore.metadata.configs.display import Field, ListDisplay
from wbcore.metadata.configs.display.instance_display import Display
from wbcore.metadata.configs.display.instance_display.shortcuts import (
    create_simple_display,
)
from wbcore.metadata.configs.display.view_config import DisplayViewConfig


class NotificationTypeSettingDisplayConfig(DisplayViewConfig):
    def get_list_display(self) -> ListDisplay:
        return ListDisplay(
            fields=[
                Field(key="notification_type", label="Notification"),
                Field(key="help_text", label="Help Text"),
                Field(key="enable_web", label="Web"),
                Field(key="enable_mobile", label="Mobile"),
                Field(key="enable_email", label="E-Mail"),
            ],
        )

    def get_instance_display(self) -> Display:
        return create_simple_display(
            [
                ["notification_type", "notification_type", "notification_type"],
                ["enable_web", "enable_mobile", "enable_email"],
            ]
        )



from django.shortcuts import get_object_or_404
from django.utils.translation import gettext as _
from wbcore.contrib.example_app.models import Match
from wbcore.metadata.configs import display as dp
from wbcore.metadata.configs.display import (
    Display,
    Inline,
    Layout,
    Page,
    Section,
    Style,
)
from wbcore.metadata.configs.display.instance_display.operators import default, lte
from wbcore.metadata.configs.display.list_display import ListDisplay
from wbcore.metadata.configs.display.view_config import DisplayViewConfig

MATCH_FIELDS = [
    dp.Field(key="home", label=_("Home")),
    dp.Field(key="score_home", label=_("Home Score")),
    dp.Field(key="away", label=_("Away")),
    dp.Field(key="score_away", label=_("Away Score")),
    dp.Field(key="league", label=_("League")),
    dp.Field(key="date_time", label=_("Date Time")),
    dp.Field(key="stadium", label=_("Stadium")),
    dp.Field(key="referee", label=_("Referee")),
]


def get_match_status_legend() -> list[dp.Legend]:
    """Dynamically creates the match legend based on the status enum using the color mapping"""

    legend_items = []
    for status, color in Match.MatchStatus.get_color_map():
        legend_items.append(dp.LegendItem(icon=color, label=status.label, value=status.value))
    return [dp.Legend(key="status", items=legend_items)]


def get_match_status_formatting() -> list[dp.Formatting]:
    """Dynamically creates the match list formatting based on the status enum using the color mapping"""

    formatting_rules = []
    for status, color in Match.MatchStatus.get_color_map():
        formatting_rules.append(dp.FormattingRule(condition=("==", status.value), style={"backgroundColor": color}))
    return [dp.Formatting(column="status", formatting_rules=formatting_rules)]


class MatchDisplayConfig(DisplayViewConfig):
    def get_list_display(self) -> ListDisplay:
        return dp.ListDisplay(
            fields=MATCH_FIELDS,
            legends=get_match_status_legend(),
            formatting=get_match_status_formatting(),
        )

    def get_instance_display(self) -> Display:
        events_section = Section(
            key="events_section",
            collapsible=False,
            title=_("Events"),
            display=Display(
                pages=[
                    Page(
                        title=_("Events"),
                        layouts={
                            default(): Layout(
                                grid_template_areas=[["events_inline"]],
                                grid_template_columns=[
                                    "minmax(min-content, 1fr)",
                                ],
                                grid_auto_rows=Style.MIN_CONTENT,
                                inlines=[Inline(key="events_inline", endpoint="event_list")],
                            )
                        },
                    ),
                ]
            ),
        )
        ongoing_or_finished_display = Display(
            pages=[
                Page(
                    title=_("Score"),
                    layouts={
                        default(): Layout(
                            grid_template_areas=[
                                ["score_home", "score_away"],
                                ["home", "away"],
                                ["events_section", "events_section"],
                            ],
                            grid_auto_columns="minmax(min-content, 1fr)",
                            grid_auto_rows=Style.MIN_CONTENT,
                            column_gap=Style.rem(5),
                            sections=[events_section],
                        ),
                    },
                ),
                Page(
                    title=_("Match Details"),
                    layouts={
                        default(): Layout(
                            grid_template_areas=[
                                ["date_time", "stadium", "referee"],
                                ["sport", "league", "."],
                            ],
                            grid_auto_columns="minmax(min-content, 1fr)",
                            grid_auto_rows=Style.MIN_CONTENT,
                            column_gap=Style.rem(5),
                        ),
                        lte(900): Layout(
                            grid_template_areas=[
                                ["date_time", "stadium"],
                                ["referee", "sport"],
                                ["league", "."],
                            ],
                            grid_auto_columns="minmax(min-content, 1fr)",
                            grid_auto_rows=Style.MIN_CONTENT,
                            column_gap=Style.rem(5),
                        ),
                        lte(400): Layout(
                            grid_template_areas=[["date_time"], ["stadium"], ["referee"], ["sport"], ["league"]],
                            grid_auto_columns="minmax(min-content, 1fr)",
                            grid_auto_rows=Style.MIN_CONTENT,
                        ),
                    },
                ),
            ]
        )
        scheduled_display = Display(
            pages=[
                Page(
                    layouts={
                        default(): Layout(
                            grid_template_areas=[
                                ["home", "away", "referee"],
                                ["date_time", "stadium", "."],
                                ["sport", "league", "."],
                            ],
                            grid_auto_columns="minmax(min-content, 1fr)",
                            grid_auto_rows=Style.MIN_CONTENT,
                            column_gap=Style.rem(5),
                        ),
                        lte(900): Layout(
                            grid_template_areas=[
                                ["home", "away"],
                                ["date_time", "stadium"],
                                ["sport", "league"],
                                ["referee", "."],
                            ],
                            grid_auto_columns="minmax(min-content, 1fr)",
                            grid_auto_rows=Style.MIN_CONTENT,
                            column_gap=Style.rem(3),
                        ),
                        lte(550): Layout(
                            grid_template_areas=[
                                ["home"],
                                ["away"],
                                ["date_time"],
                                ["stadium"],
                                ["sport"],
                                ["league"],
                                ["referee"],
                            ],
                            grid_auto_columns="minmax(min-content, 1fr)",
                            grid_auto_rows=Style.MIN_CONTENT,
                        ),
                    },
                ),
            ]
        )

        if match_id := self.view.kwargs.get("pk"):
            # Can this be optimized in any way?
            match_obj = get_object_or_404(Match, pk=match_id)
            if match_obj.status != Match.MatchStatus.SCHEDULED:
                return ongoing_or_finished_display
        return scheduled_display


class MatchStadiumDisplayConfig(MatchDisplayConfig):
    def get_list_display(self) -> ListDisplay:
        fields = MATCH_FIELDS.copy()
        fields.pop(6)
        return dp.ListDisplay(
            fields=fields,
            legends=get_match_status_legend(),
            formatting=get_match_status_formatting(),
        )

    def get_instance_display(self) -> Display:
        if "pk" in self.view.kwargs:
            return super().get_instance_display()

        return Display(
            pages=[
                Page(
                    layouts={
                        default(): Layout(
                            grid_template_areas=[["home", "away", "date_time"], ["league", "referee", "."]],
                            grid_auto_columns="minmax(min-content, 1fr)",
                            grid_auto_rows=Style.MIN_CONTENT,
                            column_gap=Style.rem(5),
                        ),
                        lte(900): Layout(
                            grid_template_areas=[["home", "away"], ["league", "referee"], ["date_time", "."]],
                            grid_auto_columns="minmax(min-content, 1fr)",
                            grid_auto_rows=Style.MIN_CONTENT,
                            column_gap=Style.rem(3),
                        ),
                        lte(600): Layout(
                            grid_template_areas=[["home"], ["away"], ["league"], ["date_time"], ["referee"]],
                            grid_auto_columns="minmax(min-content, 1fr)",
                            grid_auto_rows=Style.MIN_CONTENT,
                        ),
                    },
                ),
            ]
        )


class MatchLeagueDisplayConfig(MatchDisplayConfig):
    def get_list_display(self) -> ListDisplay:
        fields = MATCH_FIELDS.copy()
        fields.pop(4)
        return dp.ListDisplay(
            fields=fields,
            legends=get_match_status_legend(),
            formatting=get_match_status_formatting(),
        )

    def get_instance_display(self) -> Display:
        if "pk" in self.view.kwargs:
            return super().get_instance_display()

        return Display(
            pages=[
                Page(
                    layouts={
                        default(): Layout(
                            grid_template_areas=[["home", "away", "sport"], ["stadium", "date_time", "referee"]],
                            grid_auto_columns="minmax(min-content, 1fr)",
                            grid_auto_rows=Style.MIN_CONTENT,
                            column_gap=Style.rem(5),
                        ),
                        lte(900): Layout(
                            grid_template_areas=[["home", "away"], ["stadium", "date_time"], ["sport", "referee"]],
                            grid_auto_columns="minmax(min-content, 1fr)",
                            grid_auto_rows=Style.MIN_CONTENT,
                            column_gap=Style.rem(3),
                        ),
                        lte(600): Layout(
                            grid_template_areas=[
                                ["home"],
                                ["away"],
                                ["stadium"],
                                ["date_time"],
                                ["sport"],
                                ["referee"],
                            ],
                            grid_auto_columns="minmax(min-content, 1fr)",
                            grid_auto_rows=Style.MIN_CONTENT,
                        ),
                    },
                ),
            ]
        )



from django.contrib import admin
from django.contrib.contenttypes.models import ContentType


@admin.register(ContentType)
class ContentTypeModelAdmin(admin.ModelAdmin):
    list_display = ["app_label", "model"]
    search_fields = ["app_label", "model"]



from django.contrib.messages import warning
from django.db.models import Q, QuerySet
from django.utils.translation import gettext as _
from wbcore import viewsets
from wbcore.contrib.example_app.filters import (
    MatchFilter,
    MatchLeagueFilter,
    MatchStadiumFilter,
)
from wbcore.contrib.example_app.models import Match
from wbcore.contrib.example_app.serializers import (
    MatchModelSerializer,
    MatchRepresentationSerializer,
    ReadOnlyMatchModelSerializer,
)
from wbcore.contrib.example_app.viewsets.displays import (
    MatchDisplayConfig,
    MatchLeagueDisplayConfig,
    MatchStadiumDisplayConfig,
)
from wbcore.contrib.example_app.viewsets.endpoints import (
    MatchLeagueEndpointConfig,
    MatchStadiumEndpointConfig,
)
from wbcore.contrib.example_app.viewsets.titles import (
    MatchLeagueTitleConfig,
    MatchStadiumTitleConfig,
    MatchTitleConfig,
)


class MatchModelViewSet(viewsets.ModelViewSet):
    display_config_class = MatchDisplayConfig
    queryset = Match.objects.all()
    serializer_class = MatchModelSerializer
    title_config_class = MatchTitleConfig
    search_fields = ("home__name", "away__name", "league__computed_str", "sport__name")
    ordering = ("-date_time", "home__name", "away__name")
    ordering_fields = (
        "home__name",
        "away__name",
        "league__computed_str",
        "date_time",
        "stadium__name",
        "score_home",
        "score_away",
        "referee__computed_str",
        "sport__name",
    )
    filterset_class = MatchFilter

    def get_serializer_class(self):
        if "pk" in self.kwargs:
            if match_obj := self.get_object():
                if match_obj.status != Match.MatchStatus.SCHEDULED:
                    return ReadOnlyMatchModelSerializer
        return super().get_serializer_class()

    def add_messages(self, request, instance: Match | None = None, **kwargs):
        if instance and (home_stadium := instance.home.home_stadium) and instance.stadium.pk != home_stadium.pk:
            warning(request, _("Home team's home stadium not selected!"))

    def get_queryset(self) -> QuerySet[Match]:
        queryset = super().get_queryset().select_related("home", "away", "stadium", "referee", "league", "sport")
        if team_id := self.kwargs.get("team_id"):
            queryset = queryset.filter(Q(home=team_id) | Q(away=team_id))
        return queryset


class MatchStadiumModelViewSet(MatchModelViewSet):
    display_config_class = MatchStadiumDisplayConfig
    ordering_fields = (
        "home__name",
        "away__name",
        "date_time",
        "score_home",
        "score_away",
        "referee__computed_str",
        "league__computed_str",
        "sport__name",
    )
    filterset_class = MatchStadiumFilter
    endpoint_config_class = MatchStadiumEndpointConfig
    title_config_class = MatchStadiumTitleConfig

    def get_queryset(self) -> QuerySet[Match]:
        queryset = super().get_queryset()
        if stadium_id := self.kwargs.get("stadium_id"):
            queryset = queryset.filter(stadium=stadium_id)
        return queryset


class MatchLeagueModelViewSet(MatchModelViewSet):
    display_config_class = MatchLeagueDisplayConfig
    ordering_fields = (
        "home__name",
        "away__name",
        "date_time",
        "score_home",
        "score_away",
        "referee__computed_str",
        "stadium__name",
        "sport__name",
    )
    filterset_class = MatchLeagueFilter
    endpoint_config_class = MatchLeagueEndpointConfig
    title_config_class = MatchLeagueTitleConfig

    def get_queryset(self) -> QuerySet[Match]:
        queryset = super().get_queryset()
        if league_id := self.kwargs.get("league_id"):
            queryset = queryset.filter(league=league_id)
        return queryset


class MatchRepresentationViewSet(viewsets.RepresentationViewSet):
    queryset = Match.objects.all()
    serializer_class = MatchRepresentationSerializer
    search_fields = ("computed_str",)



from django.utils.translation import gettext as _
from rest_framework.exceptions import ValidationError
from wbcore import serializers as wb_serializers
from wbcore.contrib.workflow.models import Condition, Transition
from wbcore.contrib.workflow.serializers.transition import (
    TransitionRepresentationSerializer,
)


class ConditionModelSerializer(wb_serializers.ModelSerializer):
    transition = wb_serializers.PrimaryKeyRelatedField(
        default=wb_serializers.DefaultFromGET("transition"),
        queryset=Transition.objects.all(),
        label=_("Transition"),
    )
    _transition = TransitionRepresentationSerializer(source="transition")

    def validate(self, data: dict) -> dict:
        operator: str | None = data.get("operator", self.instance.operator if self.instance else None)
        expected_value: str | None = data.get(
            "expected_value", self.instance.expected_value if self.instance else None
        )
        attribute_name: str | None = data.get(
            "attribute_name", self.instance.attribute_name if self.instance else None
        )
        transition: Transition | None = data.get("transition", self.instance.transition if self.instance else None)

        condition = Condition.objects.filter(
            operator=operator, expected_value=expected_value, attribute_name=attribute_name, transition=transition
        )
        if self.instance:
            condition = condition.exclude(id=self.instance.id)
        if condition.exists():
            raise ValidationError(
                {"non_field_errors": _("This condition for the selected transition already exists.")}
            )

        if transition and not (transition.to_step.workflow.model or transition.to_step.workflow.attached_data.all()):
            raise ValidationError({"transition": _("Related workflow needs to have a model or data attached.")})

        return data

    class Meta:
        model = Condition
        fields = (
            "id",
            "transition",
            "_transition",
            "attribute_name",
            "operator",
            "expected_value",
            "negate_operator",
        )


class ConditionRepresentationSerializer(wb_serializers.RepresentationSerializer):
    _detail = wb_serializers.HyperlinkField(reverse_name="wbcore:workflow:transition-detail")

    class Meta:
        model = Condition
        fields = (
            "id",
            "_detail",
        )



from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import FieldError
from wbcore import filters as wb_filters


class ContentTypeFilterSet(wb_filters.FilterSet):
    related_name_isnull = wb_filters.CharFilter(method="filter_related_name_isnull")

    def filter_related_name_isnull(self, queryset, label, value):
        if value:
            try:
                return queryset.filter(**{f"{value}__isnull": False}).distinct()
            except FieldError:
                pass
        return queryset

    class Meta:
        model = ContentType
        fields = {"id": ["in", "exact"]}



def get_columns(cursor):
    return [col[0] for col in cursor.description or []]


def dictfetchall(cursor):
    """
    Return all rows from a cursor as a dict.
    Assume the column names are unique.
    """
    columns = get_columns(cursor)
    for row in cursor.fetchall():
        yield dict(zip(columns, row))


def dictfetchone(cursor):
    columns = get_columns(cursor)
    return dict(zip(columns, cursor.fetchone() or []))



from typing import TYPE_CHECKING, Callable

from rest_framework import pagination

if TYPE_CHECKING:
    from django.db.models import QuerySet
    from rest_framework.request import Request
    from rest_framework.views import View


class EndlessPaginationMixin(pagination.BasePagination):
    """Custom pagination mixin to support disabling pagination through a query parameter."""

    def paginate_queryset(self, queryset: "QuerySet", request: "Request", view: "View | None" = None):
        """
        Paginate the queryset if pagination is not explicitly disabled.

        Args:
            queryset (QuerySet): The queryset to paginate.
            request (Request): The request object.
            view (View, optional): The view calling this method.

        Returns:
            QuerySet: The paginated queryset or None if pagination is disabled.
        """
        if request.query_params.get("disable_pagination") == "true":
            return None
        return super().paginate_queryset(queryset, request, view)


class InitialPaginationMixin:
    """Mixin to determine whether the current page is the initial page."""

    get_previous_link: Callable[[], str | None]

    def is_initial(self) -> bool:
        """
        Determine if the current page is the initial page.

        Returns:
            bool: True if it's the initial page, False otherwise.
        """
        return self.get_previous_link() is None


class CursorPagination(EndlessPaginationMixin, InitialPaginationMixin, pagination.CursorPagination):
    def _get_position_from_instance(self, instance, ordering) -> str:
        """
        Returns the super _get_position_from_instance but adjusts the first item of ordering
        to ensure that no foreignkey is included to not break the frontend.

        Args:
            instance: The instance to get the position from.
            ordering (list): The ordering criteria.

        Returns:
            Any: The position from the instance.
        """
        new_ordering = [*ordering]
        new_ordering[0] = new_ordering[0].split("__")[0]
        return super()._get_position_from_instance(instance, new_ordering)


class LimitOffsetPagination(EndlessPaginationMixin, InitialPaginationMixin, pagination.LimitOffsetPagination):
    ...



from django.contrib.contenttypes.models import ContentType
from django.utils.translation import gettext as _
from wbcore.metadata.configs.titles import TitleViewConfig


class VersionTitleConfig(TitleViewConfig):
    def get_list_title(self):
        if (content_type_id := self.view.request.GET.get("content_type", None)) and (
            object_id := self.view.request.GET.get("object_id", None)
        ):
            try:
                content_type = ContentType.objects.get(id=content_type_id)
                obj = content_type.model_class().objects.get(id=object_id)
                return _("Versions For {obj}").format(obj=str(obj))
            except Exception:
                pass
        return _("Versions")



from django.db.models import Exists, OuterRef, Q
from rest_framework.decorators import action
from rest_framework.response import Response
from wbcore import viewsets

from .buttons import ReleaseNotesButtonConfig
from .display import ReleaseNoteDisplayConfig
from .filters import ReleaseNoteFilterSet
from .models import ReleaseNote
from .serializers import ReleaseNoteModelSerializer


class ReleaseNoteReadOnlyModelViewSet(viewsets.ReadOnlyModelViewSet):
    display_config_class = ReleaseNoteDisplayConfig
    button_config_class = ReleaseNotesButtonConfig

    queryset = ReleaseNote.objects.all()
    serializer_class = ReleaseNoteModelSerializer
    filterset_class = ReleaseNoteFilterSet
    search_fields = ("module", "version", "summary")
    ordering_fields = ("module", "version", "release_date")
    ordering = ("-release_date", "-version")

    def get_queryset(self):
        return (
            super()
            .get_queryset()
            .annotate(user_read=Exists(ReleaseNote.objects.filter(id=OuterRef("id"), read_by=self.request.user)))
        )

    def retrieve(self, request, *args, **kwargs):
        obj = self.get_object()
        if not obj.read_by.filter(id=self.request.user.id).exists():
            obj.read_by.add(self.request.user)
        return super().retrieve(request, *args, **kwargs)

    @action(methods=["GET"], detail=False)
    def unread_count(self, request, pk=None):
        return Response({"count": ReleaseNote.objects.filter(~Q(read_by=self.request.user)).count()})

    @action(methods=["PATCH"], detail=False)
    def mark_all_as_read(self, request, pk=None):
        for rl in ReleaseNote.objects.filter(~Q(read_by=self.request.user)):
            rl.read_by.add(self.request.user)
        return Response({})



from django.utils.translation import gettext as _
from wbcore.contrib.example_app.models import Player, Sport
from wbcore.metadata.configs.titles import TitleViewConfig


class EventTitleConfig(TitleViewConfig):
    def get_list_title(self) -> str:
        return _("Events")

    def get_create_title(self) -> str:
        return _("Create Event")

    def get_instance_title(self) -> str:
        return _("Event")


class PlayerStatisticsChartTitleConfig(TitleViewConfig):
    def get_list_title(self) -> str:
        player_name = Player.objects.get(id=self.view.kwargs["player_id"])
        return _("Player Statistics {}").format(player_name.computed_str)


class EventTypeTitleConfig(TitleViewConfig):
    def get_list_title(self) -> str:
        return _("Event Types")

    def get_create_title(self) -> str:
        return _("Create Event Type")

    def get_instance_title(self) -> str:
        return _("Event Type")


class EventTypeSportTitleConfig(EventTypeTitleConfig):
    def get_create_title(self) -> str:
        if sport_id := self.view.kwargs.get("sport_id"):
            try:
                return _("Create {} Event Type").format(Sport.objects.get(id=sport_id).name)
            except Sport.DoesNotExist:
                pass
        return super().get_create_title()



import pytest
from wbcore.contrib.agenda.typings import ConferenceRoom as ConferenceRoomDTO


@pytest.mark.django_db
class TestDTO:
    @pytest.mark.parametrize(
        "id, name, email",
        [
            (None, None, None),
            (None, "test", None),
            (None, "test", "test@test.com"),
            (1, None, None),
            (1, "test", None),
            (1, "test", "test@test.com"),
        ],
    )
    def test_comparison(self, id, name, email):
        conf_room1 = ConferenceRoomDTO(1, "test", "test@test.com")
        conf_room2 = ConferenceRoomDTO(id=id, name=name, email=email)
        if (conf_room1.id == conf_room2.id is not None) or (conf_room1.email == conf_room2.email is not None):
            assert conf_room1 == conf_room2
        else:
            assert conf_room1 != conf_room2
        assert conf_room2 is not None



from django.contrib import admin
from wbcore.contrib.workflow.models import Condition


@admin.register(Condition)
class ConditionAdmin(admin.ModelAdmin):
    search_fields = ("name",)
    list_display = ("id", "transition", "attribute_name", "operator", "negate_operator", "expected_value")



from wbcore.contrib.workflow.sites import workflow_site


def get_model_serializer_class_for_instance(instance):
    return workflow_site.registered_model_classes_serializer_map[instance.__class__]


def get_model_serializer_class_for_class(instance_class):
    return workflow_site.registered_model_classes_serializer_map[instance_class]


def model_is_registered_for_workflow(model) -> bool:
    return model in workflow_site.registered_model_classes_serializer_map.keys()



from django.utils.translation import gettext as _
from wbcore.contrib.icons import WBIcon
from wbcore.metadata.configs import buttons as bt
from wbcore.metadata.configs.display.instance_display.shortcuts import (
    Display,
    create_simple_display,
)
from wbcore.metadata.configs.preview import PreviewViewConfig


class EntryPreviewConfig(PreviewViewConfig):
    def get_display(self) -> Display:
        fields = [
            ["computed_str"],
            ["primary_email"],
            ["primary_telephone"],
            ["primary_manager_name"],
        ]
        try:
            entry = self.view.get_object()
            if entry.profile_image:
                fields.insert(0, "profile_image")
        except Exception:
            pass

        return create_simple_display(fields)

    def get_buttons(self):
        return [
            bt.WidgetButton(key="activity", icon=WBIcon.CALENDAR.icon, label=_("Show all activities")),
        ]



from rest_framework import permissions
from rest_framework.permissions import BasePermission
from wbcore.enums import WidgetType

from .shortcuts import is_internal_user


class RestAPIModelPermissions(permissions.DjangoModelPermissions):
    """
    Mixin for adding the view permission to the perms_map
    NOTE: This is only here until Django Rest Framework patches
    their DjangoModelPermissions to include view permissions
    """

    perms_map = {
        "GET": ["%(app_label)s.view_%(model_name)s"],
        "OPTIONS": ["%(app_label)s.select_%(model_name)s"],
        "SELECT": ["%(app_label)s.select_%(model_name)s"],
        "HEAD": ["%(app_label)s.select_%(model_name)s"],
        "POST": ["%(app_label)s.add_%(model_name)s"],
        "PUT": ["%(app_label)s.change_%(model_name)s"],
        "PATCH": ["%(app_label)s.change_%(model_name)s"],
        "DELETE": ["%(app_label)s.delete_%(model_name)s"],
    }

    def has_permission(self, request, view):
        method = request.method
        if getattr(view, "WIDGET_TYPE", None) == WidgetType.SELECT.value:
            method = "SELECT"
        if getattr(view, "_ignore_model_permissions", False):
            return True

        if not request.user or (not request.user.is_authenticated and self.authenticated_users_only):
            return False

        queryset = self._queryset(view)
        perms = self.get_required_permissions(method, queryset.model)

        return request.user.has_perms(perms)


class IsInternalUser(BasePermission):
    def has_permission(self, request, view) -> bool:
        return is_internal_user(request.user, True)


class InternalUserPermissionMixin:
    def get_permissions(self):
        permissions = super().get_permissions()  # type: ignore
        return [*permissions, IsInternalUser()]



from django.urls import include, path
from wbcore.routers import WBCoreRouter

from .viewsets import viewsets

router = WBCoreRouter()
router.register(r"tagrepresentation", viewsets.TagRepresentationViewSet, basename="tagrepresentation")
router.register(r"tag", viewsets.TagModelViewSet, basename="tag")
router.register(r"taggrouprepresentation", viewsets.TagGroupRepresentationViewSet, basename="taggrouprepresentation")
router.register(r"taggroup", viewsets.TagGroupModelViewSet, basename="taggroup")

group_router = WBCoreRouter()
group_router.register(r"tag", viewsets.TagTagGroupModelViewSet, basename="group-tag")
urlpatterns = [path("", include(router.urls)), path("group/<int:group_id>/", include(group_router.urls))]



from typing import Optional

from django.utils.translation import gettext as _
from django.utils.translation import gettext_lazy
from wbcore.contrib.color.enums import WBColor
from wbcore.metadata.configs import display as dp
from wbcore.metadata.configs.display.instance_display.shortcuts import Display
from wbcore.metadata.configs.display.view_config import DisplayViewConfig


class DocumentModelDisplay(DisplayViewConfig):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        return dp.ListDisplay(
            fields=[
                dp.Field(key="name", label=_("Name")),
                dp.Field(key="document_type", label=_("Type")),
                dp.Field(key="valid_from", label=_("Valid From")),
                dp.Field(key="valid_until", label=_("Valid Until")),
                dp.Field(key="updated", label=_("Updated")),
            ],
            legends=[
                dp.Legend(
                    key="system_created",
                    items=[
                        dp.LegendItem(
                            icon=WBColor.YELLOW_LIGHT.value,
                            label=gettext_lazy("Created By System"),
                            value=True,
                        ),
                    ],
                ),
            ],
            formatting=[
                dp.Formatting(
                    column="system_created",
                    formatting_rules=[
                        dp.FormattingRule(
                            style={"backgroundColor": WBColor.YELLOW_LIGHT.value},
                            condition=("==", True),
                        ),
                    ],
                ),
            ],
        )

    def get_instance_display(self) -> Display:
        return Display(
            pages=[
                dp.Page(
                    title="Main Information",
                    layouts={
                        dp.default(): dp.Layout(
                            grid_template_areas=[
                                ["file", "file", "file", "file", "file", "file"],
                                ["name", "name", "name", "document_type", "document_type", "document_type"],
                                ["created", "created", "updated", "updated", "system_created", "system_created"],
                                [
                                    "valid_from",
                                    "valid_from",
                                    "valid_from",
                                    "valid_until",
                                    "valid_until",
                                    "valid_until",
                                ],
                                [
                                    "description",
                                    "description",
                                    "description",
                                    "description",
                                    "description",
                                    "description",
                                ],
                            ],
                            grid_template_columns=[dp.repeat_field(6, "1fr")],
                        )
                    },
                ),
                dp.Page(
                    title="Connected Items",
                    layouts={
                        dp.default(): dp.Layout(
                            grid_template_areas=[["relationships"]],
                            grid_template_columns=[dp.Style.fr(1)],
                            grid_template_rows=[dp.Style.fr(1)],
                            inlines=[dp.Inline(key="relationships", endpoint="relationships")],
                        )
                    },
                ),
                dp.Page(
                    title="Sharable Links",
                    layouts={
                        dp.default(): dp.Layout(
                            grid_template_areas=[["shareable_links"]],
                            grid_template_columns=[dp.Style.fr(1)],
                            grid_template_rows=[dp.Style.fr(1)],
                            inlines=[dp.Inline(key="shareable_links", endpoint="shareable_links")],
                        )
                    },
                ),
            ]
        )



import factory
from wbcore.contrib.notifications.models.tokens import NotificationUserToken


class NotificationUserTokenModelFactory(factory.django.DjangoModelFactory):
    user = factory.SubFactory("wbcore.contrib.authentication.factories.UserFactory")
    token = factory.Faker("pystr")
    device_type = factory.Iterator(NotificationUserToken.NotificationDeviceType.choices, getter=lambda c: c[0])

    class Meta:
        model = NotificationUserToken



import factory
from wbcore.contrib.workflow.models import Data, DataValue


class DataFactory(factory.django.DjangoModelFactory):
    label = factory.Faker("text", max_nb_chars=64)
    help_text = factory.Faker("text", max_nb_chars=128)
    data_type = factory.Iterator(Data.DataType.values)
    workflow = factory.SubFactory("wbcore.contrib.workflow.factories.WorkflowFactory")
    required = factory.Faker("pybool")

    class Meta:
        model = Data


class DataValueFactory(factory.django.DjangoModelFactory):
    value = factory.Faker("text", max_nb_chars=64)
    data = factory.SubFactory(DataFactory)
    process = factory.SubFactory("wbcore.contrib.workflow.factories.ProcessFactory")

    class Meta:
        model = DataValue



from typing import TYPE_CHECKING, Any, Iterable

from django.utils.functional import cached_property
from rest_framework.request import Request
from rest_framework.views import View

if TYPE_CHECKING:
    from wbcore.contrib.authentication.models import User
    from wbcore.contrib.directory.models import Person


class WBCoreViewConfig:
    """A class to configure a View in the WBCore application"""

    metadata_key: str
    config_class_attribute: str

    def __init__(self, view: View, request: Request, instance: bool | None = None):
        """
        Initializes an instance of WBCoreViewConfig.

        Args:
            view (View): A Django Rest Framework view.
            request (Request): An HTTP request instance.
        """
        self.view = view
        self.request = request
        self.instance = instance if instance else "pk" in view.kwargs
        self.new_mode = request.GET.get("new_mode", "false") == "true"

    @cached_property
    def user(self) -> "User":
        return self.request.user

    @cached_property
    def profile(self) -> "Person":
        return self.user.profile

    def get_metadata(self) -> Any:
        raise NotImplementedError()

    def _get_metadata(self) -> Any:
        return self.get_metadata()

    @classmethod
    def get_all_configs(cls, view: View) -> Iterable[tuple[str, "WBCoreViewConfig"]]:
        """
        Returns an iterator of tuples containing all the WBCoreViewConfig
        instances associated with the given view.

        Args:
            view (View): A Django Rest Framework view.

        Returns:
            An iterator of tuples, where each tuple contains a string with the
            name of the configuration attribute and the corresponding
            WBCoreViewConfig instance.
        """
        yield from map(
            lambda c: (c.config_class_attribute, c),
            filter(lambda x: getattr(view, x.config_class_attribute), view.config_classes),
        )

    @classmethod
    def for_view(cls, view: View, request: Request) -> "WBCoreViewConfig":
        """
        Returns the WBCoreViewConfig instance associated with the given view.

        Args:
            view (View): A Django Rest Framework view.
            request (Request): An HTTP request instance.

        Returns:
            The WBCoreViewConfig instance associated with the given view.
        """
        return getattr(view, cls.config_class_attribute)(view, request)

    @classmethod
    def as_view_mixin(cls) -> type:
        """
        Returns a mixin class that adds the configuration attribute to a view.

        Returns:
            A mixin class that adds the configuration attribute to a view.
        """
        return type(f"{cls.__name__}Mixin", (), {cls.config_class_attribute: cls})



import pytest
from wbcore.contrib.notifications.models import Notification
from wbcore.contrib.notifications.utils import base_domain


@pytest.mark.django_db
class TestNotification:
    def test_factory(self, notification: Notification):
        assert isinstance(notification, Notification)
        assert notification.pk is not None

    def test_to_str(self, notification: Notification):
        assert str(notification) == f"{notification.user} {notification.title}"

    def test_endpoint_basename(self):
        assert Notification.get_endpoint_basename() == "wbcore:notifications:notification"

    def test_representation_value_key(self):
        assert Notification.get_representation_value_key() == "id"

    def test_representation_label_key(self):
        assert Notification.get_representation_label_key() == "{{title}}"

    @pytest.mark.parametrize("notification__endpoint", ["/wbcore/notifications/"])
    def test_full_valid_internal_endpoint(self, notification):
        assert notification.get_full_endpoint() == f"{base_domain()}{notification.endpoint}"

    @pytest.mark.parametrize("notification__endpoint", ["/some_invalid_namespace/notifications/"])
    def test_full_invalid_internal_endpoint(self, notification):
        assert notification.get_full_endpoint() is None

    @pytest.mark.parametrize("notification__endpoint", ["/wbcore/notifications/"])
    def test_full_internal_endpoint_as_shareable_link(self, notification):
        assert (
            notification.get_full_endpoint(as_shareable_internal_link=True)
            == f"{base_domain()}?widget_endpoint={notification.endpoint}"
        )

    @pytest.mark.parametrize("notification__endpoint", ["https://www.google.com"])
    def test_full_valid_external_endpoint(self, notification):
        assert notification.get_full_endpoint() == notification.endpoint

    @pytest.mark.parametrize("notification__endpoint", ["https.www.google.com"])
    def test_full_invalid_external_endpoint(self, notification):
        assert notification.get_full_endpoint() is None



import operator
from functools import reduce
from typing import Any, Iterable

from pandas import DataFrame

OverwriteRule = dict[str, dict[str, str] | dict[str, Any]]


def rule(key: str, operator: str, value: str) -> dict[str, str]:
    return {
        "key": key,
        "operator": operator,
        "value": value,
    }


def overwrite(**overwrites) -> dict[str, Any]:
    return {**overwrites}


def overwrite_rule(key: str, operator: str, value: str, **overwrites) -> OverwriteRule:
    return {
        "rule": rule(key, operator, value),
        "overwrite": overwrite(**overwrites),
    }


def overwrite_row(fields: Iterable[OverwriteRule] | None = None) -> dict[str, list[OverwriteRule]]:
    overwrites_dict = {}

    if fields:
        overwrites_dict["fields"] = [*fields]

    return overwrites_dict


def overwrite_row_df(df: DataFrame, row_cond, fields: Iterable[OverwriteRule]):
    df.loc[row_cond, "_overwrites"] = [{"fields": fields}]


def override_number_to_percent(df: DataFrame, *conditions, filter_operator=operator.or_):
    reduced_conditions = reduce(filter_operator, conditions)

    df.loc[reduced_conditions, "_overwrites"] = [
        {
            "fields": [
                overwrite_rule(
                    key="type",
                    operator="=",
                    value="number",
                    decorators=[{"position": "right", "value": "%"}],
                    precision=1,
                    display_mode="decimal",
                )
            ]
        }
    ] * reduced_conditions.sum()


def override_number_with_currency(df: DataFrame, currency: str, *conditions, filter_operator=operator.or_):
    reduced_conditions = reduce(filter_operator, conditions)

    df.loc[reduced_conditions, "_overwrites"] = [
        {
            "fields": [
                overwrite_rule(
                    key="type",
                    operator="=",
                    value="number",
                    decorators=[{"position": "left", "value": currency}],
                )
            ]
        }
    ] * reduced_conditions.sum()


def override_number_to_x(df: DataFrame, *conditions, filter_operator=operator.or_):
    reduced_conditions = reduce(filter_operator, conditions)

    df.loc[reduced_conditions, "_overwrites"] = [
        {
            "fields": [
                overwrite_rule(
                    key="type",
                    operator="=",
                    value="number",
                    decorators=[{"position": "right", "value": "x"}],
                    precision=1,
                    display_mode="decimal",
                )
            ]
        }
    ] * reduced_conditions.sum()


def override_number_precision(df: DataFrame, precision=2, *conditions, filter_operator=operator.or_):
    reduced_conditions = reduce(filter_operator, conditions)

    df.loc[reduced_conditions, "_overwrites"] = [
        {
            "fields": [
                overwrite_rule(
                    key="type",
                    operator="=",
                    value="number",
                    precision=precision,
                )
            ]
        }
    ] * reduced_conditions.sum()


def override_number_to_decimal(df: DataFrame, *conditions, filter_operator=operator.or_):
    reduced_conditions = reduce(filter_operator, conditions)

    df.loc[reduced_conditions, "_overwrites"] = [
        {
            "fields": [
                overwrite_rule(
                    key="type",
                    operator="=",
                    value="number",
                    precision=0,
                    displayMode="decimal",
                )
            ]
        }
    ] * reduced_conditions.sum()



import sys
import threading
from contextlib import suppress

from ..abstract_backend import AbstractNotificationBackend


class NotificationBackend(AbstractNotificationBackend):
    @classmethod
    def send_notification(cls, notification):
        """Write the notification to the stream in a thread-safe way."""

        _lock = threading.RLock()
        stream = sys.stdout
        with _lock:
            with suppress(Exception):
                stream.write("%s\n" % notification.title)
                stream.write("%s\n" % notification.body or "")
                stream.write("-" * 79)
                stream.write("\n")
                stream.flush()  # flush after each message

    @classmethod
    def get_configuration(cls) -> dict:
        return {}



from .condition import ConditionModelViewSet
from .data import DataModelViewSet
from .process import ProcessModelViewSet, ProcessStepModelViewSet, AssignedProcessStepModelViewSet
from .step import (
    DecisionStepModelViewSet,
    DisplayModelViewSet,
    EmailStepModelViewSet,
    FinishStepModelViewSet,
    JoinStepModelViewSet,
    ScriptStepModelViewSet,
    SplitStepModelViewSet,
    StepModelViewSet,
    UserStepModelViewSet,
    StartStepModelViewSet,
)
from .transition import TransitionModelViewSet
from .workflow import WorkflowModelViewSet



from typing import Any, Dict

from django.db import models
from wbcore.contrib.io.imports import ImportExportHandler


class CurrencyImportHandler(ImportExportHandler):
    MODEL_APP_LABEL: str = "currency.Currency"

    def _deserialize(self, data: Dict[str, Any]):
        if key := data.get("key"):
            if key == "CNH":  # We switch CNH to CNY because we don't support it
                data["key"] = "CNY"
            if key == "GBp":
                data["key"] = "GBX"

    def _get_instance(self, data: Dict[str, Any], **kwargs) -> models.Model | None:
        try:
            if key := data.get("key", None):
                return self.model.objects.get(key=key)
            if _id := data.get("id", None):
                return self.model.objects.get(id=_id)
        except self.model.DoesNotExist:
            pass



from .base import WBCoreViewConfig


class SearchFieldsViewConfig(WBCoreViewConfig):
    metadata_key = "search_fields"
    config_class_attribute = "search_fields_config_class"

    def get_metadata(self) -> list[str]:
        return getattr(self.view, "search_fields", [])



# Generated by Django 4.1.8 on 2023-04-26 08:45

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("directory", "0002_auto_20230414_1553"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="entry",
            name="last_event",
        ),
    ]



import pytest
from pytest_factoryboy import register
from rest_framework.test import APIClient, APIRequestFactory
from wbcore.contrib.authentication.factories import UserFactory
from wbcore.contrib.directory.factories import PersonFactory
from wbcore.contrib.notifications.factories.notification_types import (
    NotificationTypeModelFactory,
    NotificationTypeSettingModelFactory,
)
from wbcore.contrib.notifications.factories.notifications import (
    NotificationModelFactory,
)
from wbcore.contrib.notifications.factories.tokens import (
    NotificationUserTokenModelFactory,
)
from wbcore.tests.conftest import *

register(UserFactory)
register(PersonFactory)
register(NotificationTypeModelFactory, name="notification_type")
register(NotificationTypeSettingModelFactory, name="notification_type_setting")
register(NotificationModelFactory, name="notification")
register(NotificationUserTokenModelFactory, name="notification_user_token")


@pytest.fixture
def request_with_user(user):
    request = APIRequestFactory().get("/")
    request.user = user
    return request


@pytest.fixture
def client():
    return APIClient()


@pytest.fixture(
    autouse=True, scope="session"
)  # Might want to find a way to registered default conftest logic automatically
def django_test_environment(django_test_environment):
    from django.apps import apps

    get_models = apps.get_models

    for m in [m for m in get_models() if not m._meta.managed]:
        m._meta.managed = True



from .buttons import add_object_permission_button
from .displays import PivotUserObjectPermissionDisplayViewConfig
from .endpoints import PivotUserObjectPermissionEndpointViewConfig
from .titles import PivotUserObjectPermissionTitleViewConfig



from configurations import values


class Uvicorn:
    UVICORN_LOOP = values.Value("auto", environ_prefix=None)
    # assert UVICORN_LOOP in ["auto", "uvloop", "asyncio"], "UVICORN_LOOP has an invalid setting"

    UVICORN_WS = values.Value("auto", environ_prefix=None)
    # assert UVICORN_WS in ["auto", "websockets", "wsproto"], "UVICORN_WS has an invalid setting"



from django.utils.translation import gettext as _
from wbcore.menus import ItemPermission, MenuItem
from wbcore.permissions.shortcuts import is_internal_user


def default_calendar_get_params(request) -> dict:
    return {"entities": request.user.profile.id}


CALENDAR_MENUITEM = MenuItem(
    label=_("Calendar"),
    endpoint="wbcore:agenda:calendaritem-list",
    permission=ItemPermission(
        method=lambda request: is_internal_user(request.user), permissions=["agenda.view_calendaritem"]
    ),
    endpoint_get_parameters=default_calendar_get_params,
)



from django.contrib import admin
from wbcore.contrib.agenda.models import Building, ConferenceRoom


@admin.register(ConferenceRoom)
class ConferenceRoomModelAdmin(admin.ModelAdmin):
    autocomplete_fields = ("building",)
    list_display = ["name", "email", "capacity", "building", "is_videoconference_capable"]
    search_fields = ("name", "email", "building__name")
    raw_id_fields = ["building"]


@admin.register(Building)
class BuildingModelAdmin(admin.ModelAdmin):
    autocomplete_fields = ("address",)
    list_display = ["name", "address"]
    search_fields = ("name",)
    raw_id_fields = ["address"]



from django_celery_beat.models import IntervalSchedule, PeriodicTask
from django.db import DEFAULT_DB_ALIAS
from django.apps import apps as global_apps
from django.conf import settings

from wbcore.contrib.io.models import Source


def load_sources_from_settings(
    app_config, verbosity=2, interactive=True, using=DEFAULT_DB_ALIAS, apps=global_apps, **kwargs
):
    if module_settings := getattr(settings, "DEFAULT_REGISTERED_DATA_BACKEND", None):
        Source.load_sources_from_settings(module_settings)



import pytest
from wbcore.test import GenerateTest, default_config

config = {}
for key, value in default_config.items():
    config[key] = list(
        filter(
            lambda x: x.__module__.startswith("wbcore.contrib.workflow")
            and x.__name__
            not in [
                "WorkflowModelViewSet",
                "EmailStepModelViewSet",
                "ProcessStepModelViewSet",
                "AssignedProcessStepModelViewSet",
            ],
            value,
        )
    )


@pytest.mark.django_db
@GenerateTest(config)
class TestProject:
    pass



import re

from dynamic_preferences.registries import global_preferences_registry
from wbcore.markdown.models import Asset
from weasyprint import default_url_fetcher


def custom_url_fetcher(url):
    try:
        if s := re.search(
            r".*wbcore/markdown/asset/([0-9a-fA-F]{8}\-[0-9a-fA-F]{4}\-[0-9a-fA-F]{4}\-[0-9a-fA-F]{4}\-[0-9a-fA-F]{12})",
            url,
        ):
            asset = Asset.objects.get(id=s.group(1))
            url = asset.file.url
        return default_url_fetcher(url)
    except Exception:
        global_preferences = global_preferences_registry.manager()
        default_placeholder_url = global_preferences.get("wbcore__default_empty_image_placeholder", no_cache=True)
        return default_url_fetcher(default_placeholder_url)


def remove_styled_table(html_content):
    tags = ["table", "col", "td"]
    for tag in tags:
        reg_str = "<" + tag + "(.*?)" + ">"
        reg_style = '(style=".*?;")'
        _attributs = re.findall(reg_str, html_content)
        dict_attrs = {}
        for attribut in _attributs:
            dict_attrs[attribut] = re.sub(reg_style, "", attribut)

        for _key, _value in dict_attrs.items():
            html_content = re.sub(_key, _value, html_content)

    return html_content



from wbcore import filters as wb_filters
from wbcore.filters.defaults import current_month_date_range

from .models import UserActivity


class UserActivityChartFilter(wb_filters.FilterSet):
    date = wb_filters.DateTimeRangeFilter(
        method=wb_filters.DateRangeFilter.base_date_range_filter_method,
        label="Date Range",
        required=True,
        clearable=False,
        default=current_month_date_range,
    )

    class Meta:
        model = UserActivity
        fields = {}



import factory
from django.db.models import signals
from wbcore.contrib.icons import WBIcon
from wbcore.contrib.workflow.models import Transition


@factory.django.mute_signals(signals.post_save)
class TransitionFactory(factory.django.DjangoModelFactory):
    name = factory.Faker("text", max_nb_chars=128)
    to_step = factory.SubFactory("wbcore.contrib.workflow.factories.RandomChildStepFactory")
    from_step = factory.SubFactory("wbcore.contrib.workflow.factories.RandomChildStepFactory")
    icon = factory.Iterator(WBIcon.values)

    @classmethod
    def _create(cls, model_class, *args, **kwargs):
        # Both steps need to belong to the same workflow
        if (from_step := kwargs.get("from_step")) and (to_step := kwargs.get("to_step")):
            if kwargs["from_step"].workflow != kwargs["to_step"].workflow:
                from_step.workflow = to_step.workflow
                from_step.save()
                kwargs["from_step"] = from_step
        return super()._create(model_class, *args, **kwargs)

    class Meta:
        model = Transition



from django.utils.translation import gettext_lazy as _
from wbcore import filters
from wbcore.contrib.example_app.models import League, Sport, SportPerson, Team
from wbcore.contrib.geography.models import Geography


class LeagueSportFilter(filters.FilterSet):
    established_date__gte = filters.DateFilter(
        label=_("Established Date"),
        lookup_expr="gte",
        field_name="established_date",
    )
    established_date__lte = filters.DateFilter(
        label=_("Established Date"),
        lookup_expr="lte",
        field_name="established_date",
    )
    commissioner = filters.ModelMultipleChoiceFilter(
        label=_("Commissioners"),
        queryset=SportPerson.objects.all(),
        endpoint=SportPerson.get_representation_endpoint(),
        value_key=SportPerson.get_representation_value_key(),
        label_key=SportPerson.get_representation_label_key(),
    )
    country = filters.ModelMultipleChoiceFilter(
        label=_("Countries"),
        queryset=Geography.countries.all(),
        endpoint=Geography.get_representation_endpoint(),
        value_key=Geography.get_representation_value_key(),
        label_key=Geography.get_representation_label_key(),
        filter_params={"level": 1},
    )
    teams = filters.ModelChoiceFilter(
        label=_("Team"),
        queryset=Team.objects.all(),
        endpoint=Team.get_representation_endpoint(),
        value_key=Team.get_representation_value_key(),
        label_key=Team.get_representation_label_key(),
    )
    points_per_win__gte = filters.NumberFilter(
        label=_("Points Per Win"), lookup_expr="gte", field_name="points_per_win"
    )
    points_per_win__lte = filters.NumberFilter(
        label=_("Points Per Win"), lookup_expr="lte", field_name="points_per_win"
    )
    points_per_draw__gte = filters.NumberFilter(
        label=_("Points Per Draw"), lookup_expr="gte", field_name="points_per_draw"
    )
    points_per_draw__lte = filters.NumberFilter(
        label=_("Points Per Draw"), lookup_expr="lte", field_name="points_per_draw"
    )
    points_per_loss__gte = filters.NumberFilter(
        label=_("Points Per Loss"), lookup_expr="gte", field_name="points_per_loss"
    )
    points_per_loss__lte = filters.NumberFilter(
        label=_("Points Per Loss"), lookup_expr="lte", field_name="points_per_loss"
    )

    class Meta:
        model = League
        fields = {
            "name": ["exact", "icontains"],
            "website": ["exact", "icontains"],
        }


class LeagueFilter(LeagueSportFilter):
    sport = filters.ModelMultipleChoiceFilter(
        label=_("Sports"),
        queryset=Sport.objects.all(),
        endpoint=Sport.get_representation_endpoint(),
        value_key=Sport.get_representation_value_key(),
        label_key=Sport.get_representation_label_key(),
    )



from contextlib import suppress

from django_filters.utils import get_model_field

from .lookups import get_lookup_icon, get_lookup_label


class WBCoreFilterMixin:
    def __init__(self, *args, **kwargs):
        self.default = kwargs.pop("default", None)
        self.initial = kwargs.pop("initial", None)
        self.required = kwargs.pop("required", False)
        self.clearable = kwargs.pop("clearable", True)  # TODO: Take away
        self.hidden = kwargs.pop("hidden", False)
        self.column_field_name = kwargs.pop("column_field_name", None)
        self.help_text = kwargs.pop("help_text", None)
        self.allow_empty_default = kwargs.pop("allow_empty_default", False)
        self.label_format = kwargs.pop(
            "label_format",
            getattr(self, "default_label_format", "{{field_label}} {{operation_icon}}  {{value_label}}"),
        )
        self.lookup_icon = kwargs.pop("lookup_icon", None)
        self.lookup_label = kwargs.pop("lookup_label", None)
        self.depends_on = kwargs.pop("depends_on", [])
        super().__init__(*args, **kwargs)

    @property
    def key(self):
        return self.column_field_name if self.column_field_name else self.field_name

    def get_label(self):
        if self.label is None:  # if label is not provided we gracefully convert the field name into capitalized label
            return self.field_name.replace("_", " ").title()
        else:
            return self.label

    def _get_default(self, request, view):
        # We consider the case where default is a boolean with value False.
        if callable(self.default):
            default = self.default(self, request, view)
        elif (
            isinstance(self.default, str)
            and (callable_default := getattr(self, self.default, None))
            and (callable(callable_default))
        ):
            default = callable_default(self, request, view)
        else:
            default = self.default

        return default

    def _validate_default_with_request(self, default, request, name):
        if request_default := request.GET.get(name):
            return request_default
        return default

    def get_help_text(self) -> str:
        if self.help_text:
            return self.help_text
        with suppress(AttributeError):
            field = get_model_field(self.parent._meta.model, self.field_name)
            if field.help_text:
                return field.help_text
        if self.label:
            return "Filter by " + self.label

    def get_representation(self, request, name, view):
        if self.hidden:
            return {}

        representation = {
            "key": self.key,
            "label_format": self.label_format,
            "label": self.get_label(),
            "help_text": self.get_help_text(),
        }
        lookup_expr = {
            "label": get_lookup_label(self.lookup_expr) if self.lookup_label is None else self.lookup_label,
            "icon": get_lookup_icon(self.lookup_expr) if self.lookup_icon is None else self.lookup_icon,
            "key": name,
            "hidden": self.hidden,
            "input_properties": {
                "type": self.filter_type,
            },
        }
        default = self._get_default(request, view)
        if default is not None or self.allow_empty_default:
            lookup_expr["input_properties"]["default"] = default

        initial = None
        if callable(self.initial):
            initial = self.initial(self, request, view)
        elif (
            isinstance(self.initial, str)
            and (callable_initial := getattr(self, self.initial, None))
            and (callable(callable_initial))
        ):
            initial = callable_initial(self, request, view)
        else:
            initial = self.initial

        if _initial := self._validate_default_with_request(initial, request, name):
            initial = _initial

        if initial is not None:
            lookup_expr["input_properties"]["initial"] = initial

        if self.required:
            lookup_expr["input_properties"]["required"] = True
            # assert representation["default"] != {}, "If a filter is required, it needs at least one default value"
        representation["depends_on"] = self.depends_on
        return representation, lookup_expr



from datetime import timedelta

from django.utils.translation import gettext_lazy as _
from wbcore import filters as wb_filters
from wbcore.contrib.agenda.models import CalendarItem, ConferenceRoom
from wbcore.contrib.directory.models import Entry
from wbcore.utils.date import current_month_date_end, current_month_date_start


class CalendarItemPeriodBaseFilterSet(wb_filters.FilterSet):
    period = wb_filters.DateTimeRangeFilter(label="Period")

    conference_room = wb_filters.ModelMultipleChoiceFilter(
        label=_("Conference Rooms"),
        queryset=ConferenceRoom.objects.all(),
        endpoint=ConferenceRoom.get_representation_endpoint(),
        value_key=ConferenceRoom.get_representation_value_key(),
        label_key=ConferenceRoom.get_representation_label_key(),
    )
    boolean_conference_room = wb_filters.BooleanFilter(
        label=_("Inside Conference Room"), method="boolean_is_in_conference_room"
    )

    def boolean_is_in_conference_room(self, queryset, name, value):
        if value:
            return queryset.filter(conference_room__isnull=False)
        elif value is False:
            return queryset.filter(conference_room__isnull=True)
        return queryset

    class Meta:
        model = CalendarItem
        fields = {}


def get_calendar_period_default(*args, **kwargs) -> dict:
    month_start = current_month_date_start(*args, **kwargs)
    last_monday_from_month_start = month_start - timedelta(days=month_start.weekday())

    week_after_month = current_month_date_end(*args, **kwargs) + timedelta(weeks=1)
    next_week_sunday_after_month = week_after_month + timedelta(
        days=6 - week_after_month.weekday()
    )  # weekday will return value between 0 and 6 so in order to get the next Sunday we need to subtract from 6

    return {"period": f"{last_monday_from_month_start:%Y-%m-%d},{next_week_sunday_after_month:%Y-%m-%d}"}


class CalendarItemFilter(CalendarItemPeriodBaseFilterSet):
    period = wb_filters.DateTimeRangeFilter(label="Period", required=True, default=get_calendar_period_default)
    item_type = wb_filters.MultipleChoiceFilter(
        label=_("Type"), choices=CalendarItem.get_item_types_choices(), field_name="item_type"
    )
    entities = wb_filters.ModelMultipleChoiceFilter(
        label=_("Calendar Regarding"),
        queryset=Entry.objects.all(),
        endpoint=Entry.get_representation_endpoint(),
        value_key=Entry.get_representation_value_key(),
        label_key=Entry.get_representation_label_key(),
        field_name="entities",
    )

    class Meta:
        model = CalendarItem
        fields = {}



from typing import Any
from unittest.mock import patch

import pytest
from django.contrib.contenttypes.models import ContentType
from faker import Faker
from wbcore.contrib.io.factories import ImportModel

from ..factories import ImportExportModelHandler
from ..models import Source

fake = Faker()


@pytest.fixture()
def handler(import_source):
    return ImportExportModelHandler(import_source)


def to_dict(model):
    many = []
    if model.many_relationships:
        many = [m for m in model.many_relationships.all()]

    return {
        "relationship": model.relationship,
        "import_source": model.import_source,
        "many_relationships": many,
        "number": model.number,
        "text": model.text,
        "name": model.name,
    }


@pytest.mark.django_db
class TestImportSourceModel:
    def _get_random_data(self, parser_handler_factory, relationship_id=None, object_id=None, history=False):
        res: dict[str, Any] = {
            "data": [
                {
                    "relationship": parser_handler_factory.create().id if not relationship_id else relationship_id,
                    "many_relationships": [parser_handler_factory.create().id],
                    "number": fake.pyfloat(),
                    "text": fake.paragraph(),
                    "name": fake.name(),
                    "id": object_id,
                }
            ]
        }
        if history:
            res["history"] = {"relationship": relationship_id if relationship_id else res["data"][0]["relationship"]}
        return res

    def test_init(self, import_model):
        assert import_model.id is not None

    @patch.object(Source, "generate_import_sources")
    @pytest.mark.parametrize(
        "parser_handler__handler,data_backend__passive_only",
        [
            ("wbcore.contrib.io.ImportModel", True),  # Trigger import_source
            ("wbcore.contrib.io.ImportModel", False),  # Do nothing
            ("otherapplabel.OtherModel", True),  # Do nothing
        ],
    )
    def test_import_data(self, mock_process_source, import_model, source, parser_handler, data_backend):
        source.parser_handler.add(parser_handler)
        source.save()
        content_type = ContentType.objects.get_for_model(ImportModel)
        model_name = f"{content_type.app_label}.{content_type.model}"
        nb_valid_sources = Source.objects.filter(parser_handler__handler__iexact=model_name)

        import_model.import_data()
        assert mock_process_source.call_count == nb_valid_sources.count()

    def test_model_dict_diff(self, import_model_factory, handler, import_source, parser_handler_factory):
        a11 = parser_handler_factory.create()
        a12 = parser_handler_factory.create()
        a2 = parser_handler_factory.create()
        model = import_model_factory.create(many_relationships=[a11, a12])
        comparison_model = import_model_factory.create(many_relationships=[a11, a2])

        diff_dict = handler._model_dict_diff(model, to_dict(comparison_model))
        assert diff_dict["relationship"] == comparison_model.relationship
        assert diff_dict["import_source"] == comparison_model.import_source
        assert set(diff_dict["many_relationships"]) == set([a2])
        assert diff_dict["number"] == pytest.approx(comparison_model.number, rel=10 - 6)
        assert diff_dict["text"] == comparison_model.text
        assert diff_dict["name"] == comparison_model.name

    def test_model_dict_diff_no_change(self, import_model_factory, handler, import_source, parser_handler_factory):
        a11 = parser_handler_factory.create()
        a12 = parser_handler_factory.create()
        model = import_model_factory.create(many_relationships=[a11, a12])
        comparison_model = model

        diff_dict = handler._model_dict_diff(model, to_dict(comparison_model))
        assert not diff_dict

    def test_data_changed(self, import_model_factory, handler, import_source, parser_handler_factory):
        a11 = parser_handler_factory.create()
        a12 = parser_handler_factory.create()
        a2 = parser_handler_factory.create()
        model = import_model_factory.create(many_relationships=[a11, a12], import_source=handler.import_source)
        comparison_model = import_model_factory.create(
            many_relationships=[a11, a2], import_source=handler.import_source
        )
        data = to_dict(comparison_model)
        diff_dict = handler._model_dict_diff(model, data)
        handler._data_changed(model, diff_dict, data)
        handler._data_changed(model, diff_dict, data)
        assert model.relationship == comparison_model.relationship
        assert model.import_source == comparison_model.import_source
        assert set(model.many_relationships.values_list("id", flat=True)) == set({a11.id, a12.id, a2.id})
        assert model.number == pytest.approx(comparison_model.number, rel=10 - 3)
        assert model.text == comparison_model.text
        assert model.name == comparison_model.name

    def test_data_change_only_fields(self, handler, import_source, import_model_factory):
        model = import_model_factory.create()
        comparison_model = import_model_factory.create()
        data = to_dict(comparison_model)
        diff_dict = handler._model_dict_diff(model, data)
        handler._data_changed(model, diff_dict, data, include_update_fields=["number", "text"])
        assert model.number == pytest.approx(comparison_model.number, rel=10 - 3)
        assert model.text == comparison_model.text
        assert model.name != comparison_model.name

    def test_process_wrongly_formatted_import_data(self, handler, import_source):
        with pytest.raises(Exception):
            handler.process(dict(a=1, b="b"))

    def test_process_basic(self, handler, import_source, parser_handler_factory):
        # Create basic data and call process. We expect only one object to be created
        import_data = self._get_random_data(parser_handler_factory)
        handler.process(import_data)
        created_obj = ImportModel.objects.first()
        if not created_obj:
            raise ValueError("We expect a created object")
        assert created_obj.text == import_data["data"][0]["text"]
        assert created_obj.name == import_data["data"][0]["name"]
        assert created_obj.number == import_data["data"][0]["number"]

        # If we recall process without specifying lookup identifier, we expect a second object to be created
        handler.process(import_data)
        assert ImportModel.objects.count() == 2

        # If we generate completely new data but with a lookup pointing to the first created object, we expect its value
        # to be updated instead of created.
        new_import_data = self._get_random_data(parser_handler_factory, object_id=created_obj.pk)
        handler.process(new_import_data)
        assert ImportModel.objects.count() == 2
        created_obj.refresh_from_db()
        assert created_obj.text == new_import_data["data"][0]["text"]
        assert created_obj.name == new_import_data["data"][0]["name"]
        assert created_obj.number == new_import_data["data"][0]["number"]

        import_historical_data = self._get_random_data(
            parser_handler_factory,
            object_id=created_obj.pk,
            history=True,
            relationship_id=import_data["data"][0]["relationship"],
        )
        # Even though id is provided, we expect the history to not include this object (because get_history will return
        # a queryset without it
        handler.process(import_historical_data)
        assert ImportModel.objects.count() == 3



from django.utils.translation import gettext as _
from rest_framework.reverse import reverse
from wbcore.metadata.configs import buttons as bt
from wbcore.metadata.configs.buttons.view_config import ButtonViewConfig


class BuildingButtonConfig(ButtonViewConfig):
    def get_custom_buttons(self):
        if not self.view.kwargs.get("pk", None):
            base_url = reverse("wbcore:directory:addresscontact-list", args=[], request=self.request)

            return {
                bt.WidgetButton(
                    endpoint=base_url,
                    label=_("New Address"),
                    icon="add_location",
                    new_mode=True,
                )
            }
        return {}



from django.db.models.query import QuerySet
from django.utils import timezone
from rest_framework import status
from rest_framework.decorators import action
from rest_framework.request import Request
from rest_framework.response import Response
from wbcore import viewsets
from wbcore.contrib.notifications.models import Notification
from wbcore.contrib.notifications.serializers import NotificationModelSerializer
from wbcore.contrib.notifications.viewsets.configs.notifications import (
    NotificationButtonConfig,
    NotificationDisplayConfig,
)


class NotificationModelViewSet(viewsets.ModelViewSet):
    ordering = ("-created",)
    search_fields = ("title", "notification_type__title")

    queryset = Notification.objects.all()

    serializer_class = NotificationModelSerializer
    display_config_class = NotificationDisplayConfig
    button_config_class = NotificationButtonConfig

    def retrieve(self, request: Request, *args, **kwargs) -> Response:
        obj = self.get_object()
        if not obj.read:
            obj.read = timezone.now()
            obj.save()
        return super().retrieve(request, *args, **kwargs)

    def create(self, request: Request) -> Response:
        return Response(status=status.HTTP_403_FORBIDDEN)

    def update(self, request: Request, pk=None) -> Response:
        return Response(status=status.HTTP_403_FORBIDDEN)

    def partial_update(self, request: Request, pk=None) -> Response:
        return Response(status=status.HTTP_403_FORBIDDEN)

    def get_queryset(self) -> QuerySet[Notification]:
        return super().get_queryset().filter(user=self.request.user)

    @action(methods=["GET"], detail=False)
    def unread_count(self, request, pk=None):
        return Response({"count": Notification.objects.filter(user=request.user, read__isnull=True).count()})

    @action(methods=["PATCH"], detail=False)
    def read_all(self, request: Request, pk=None) -> Response:
        self.get_queryset().filter(read__isnull=True).update(read=timezone.now())
        return Response({})

    @action(methods=["PATCH"], detail=False)
    def delete_all_read(self, request: Request, pk=None) -> Response:
        self.get_queryset().filter(read__isnull=False).delete()
        return Response({})



import pytest
from django.utils import timezone
from rest_framework.reverse import reverse
from wbcore.contrib.authentication.factories import UserFactory
from wbcore.contrib.notifications.factories.notifications import (
    NotificationModelFactory,
)
from wbcore.contrib.notifications.models.notifications import Notification


@pytest.mark.django_db
class TestNotificationModelViewSet:
    @pytest.mark.parametrize(
        "user__user_permissions,status_code", [(None, 403), (["notifications.view_notification"], 200)]
    )
    def test_instance(self, notification, client, user, status_code):
        client.force_authenticate(user)
        response = client.get(reverse("wbcore:notifications:notification-detail", args=[notification.pk]))
        assert response.status_code == status_code  # type: ignore

    @pytest.mark.parametrize("user__user_permissions,status_code", [(["notifications.view_notification"], 200)])
    def test_instance_read(self, notification, client, user, status_code):
        assert notification.read is None
        client.force_authenticate(user)
        response = client.get(reverse("wbcore:notifications:notification-detail", args=[notification.pk]))
        assert response.status_code == status_code  # type: ignore
        notification.refresh_from_db()
        assert notification.read is not None

    @pytest.mark.parametrize(
        "user__user_permissions,status_code", [(None, 403), (["notifications.view_notification"], 200)]
    )
    def test_list(self, client, user, status_code):
        client.force_authenticate(user)
        response = client.get(reverse("wbcore:notifications:notification-list"))
        assert response.status_code == status_code  # type: ignore

    @pytest.mark.parametrize("user__user_permissions,status_code", [(["notifications.add_notification"], 403)])
    def test_create(self, notification, client, user, status_code):
        """Regardless of permission, no user should ever be able to create notification"""
        client.force_authenticate(user)
        response = client.post(
            reverse("wbcore:notifications:notification-list"),
            {"user": user.id, "notification": notification.id, "title": "abc", "body": "abc"},
        )
        assert response.status_code == status_code  # type: ignore

    @pytest.mark.parametrize("user__user_permissions,status_code", [(["notifications.change_notification"], 403)])
    def test_change(self, notification, client, user, status_code):
        """Regardless of permission, no user should ever be able to change notification"""
        client.force_authenticate(user)
        response = client.patch(
            reverse("wbcore:notifications:notification-detail", args=[notification.pk]), {"title": "abc"}
        )
        assert response.status_code == status_code  # type: ignore
        response = client.put(
            reverse("wbcore:notifications:notification-detail", args=[notification.pk]), {"title": "abc"}
        )
        assert response.status_code == status_code  # type: ignore

    @pytest.mark.parametrize(
        "user__user_permissions,status_code", [(None, 403), (["notifications.delete_notification"], 204)]
    )
    def test_delete(self, notification, client, user, status_code):
        client.force_authenticate(user)
        response = client.delete(reverse("wbcore:notifications:notification-detail", args=[notification.pk]))
        assert response.status_code == status_code  # type: ignore

    @pytest.mark.parametrize("user__user_permissions,status_code", [(["notifications.delete_notification"], 404)])
    def test_delete_other_notification(self, notification, client, user, status_code):
        """Regardless of the permission, a user should never be able to delete another users notifications"""
        user2 = UserFactory()
        notification.user = user2
        notification.save()
        client.force_authenticate(user)
        response = client.delete(reverse("wbcore:notifications:notification-detail", args=[notification.pk]))
        assert response.status_code == status_code  # type: ignore

    @pytest.mark.parametrize("user__user_permissions", [(["notifications.change_notification"])])
    def test_read_all_action(self, notification, client, user):
        assert notification.read is None
        client.force_authenticate(user)
        client.patch(reverse("wbcore:notifications:notification-read-all"))
        notification.refresh_from_db()
        assert notification.read is not None

    @pytest.mark.parametrize("user__user_permissions", [(["notifications.change_notification"])])
    def test_read_all_action_other_notification(self, notification, client, user):
        notification2 = NotificationModelFactory()
        assert notification.read is None
        assert notification2.read is None  # type: ignore
        client.force_authenticate(user)
        client.patch(reverse("wbcore:notifications:notification-read-all"))
        notification.refresh_from_db()
        notification2.refresh_from_db()  # type: ignore
        assert notification.read is not None
        assert notification2.read is None  # type: ignore

    @pytest.mark.parametrize("user__user_permissions", [(["notifications.change_notification"])])
    def test_delete_all_action(self, notification, client, user):
        NotificationModelFactory(user=user, read=timezone.now())
        assert Notification.objects.filter(user=user).count() == 2

        client.force_authenticate(user)
        client.patch(reverse("wbcore:notifications:notification-delete-all-read"))

        assert Notification.objects.filter(user=user).count() == 1

    @pytest.mark.parametrize("user__user_permissions", [(["notifications.change_notification"])])
    def test_delete_all_action_other_notification(self, notification, client, user):
        user2 = UserFactory()
        NotificationModelFactory(user=user2, read=timezone.now())
        notification.read = timezone.now()
        notification.save()
        assert Notification.objects.all().count() == 2

        client.force_authenticate(user)
        client.patch(reverse("wbcore:notifications:notification-delete-all-read"))

        assert Notification.objects.all().count() == 1



from datetime import datetime

from markdown import Markdown


def parse_release_note(release_note):
    md = Markdown(extensions=["meta"], tab_length=2)
    content = md.convert(release_note)
    meta = md.Meta
    version = meta["version"][0]
    module = meta["module"][0]
    summary = meta["summary"][0]
    release_date = datetime.strptime(meta["date"][0], "%Y-%m-%d")
    return content, meta, version, module, summary, release_date



import factory

from ..models import (
    AddressContact,
    BankingContact,
    EmailContact,
    SocialMediaContact,
    TelephoneContact,
    WebsiteContact,
)


class ContactBaseFactory(factory.django.DjangoModelFactory):
    class Meta:
        abstract = True

    primary = factory.Faker("pybool")
    # location = factory.SubFactory('directory.factories.contacts.ContactLocationFactory')
    entry = factory.SubFactory("wbcore.contrib.directory.factories.entries.PersonFactory")


class BankingContactFactory(ContactBaseFactory):
    class Meta:
        model = BankingContact

    institute = factory.Faker("company")
    institute_additional = factory.Faker("sentence", nb_words=3)
    iban = factory.Faker("iban")
    swift_bic = factory.Faker("swift")
    currency = factory.SubFactory("wbcore.contrib.currency.factories.CurrencyFactory")
    additional_information = factory.Faker("paragraph")


class AddressContactFactory(ContactBaseFactory):
    class Meta:
        model = AddressContact

    street = factory.Faker("street_address")
    street_additional = factory.Faker("sentence", nb_words=3)
    zip = factory.Faker("zipcode")
    geography_city = factory.SubFactory("wbcore.contrib.geography.factories.CityFactory")


class TelephoneContactFactory(ContactBaseFactory):
    class Meta:
        model = TelephoneContact

    number = factory.Faker("phone_number")
    # telephone_type = factory.SubFactory('directory.factories.contacts.ContactTelephoneTypeFactory')


class EmailContactFactory(ContactBaseFactory):
    class Meta:
        model = EmailContact

    address = factory.Faker("email")


class WebsiteContactFactory(ContactBaseFactory):
    class Meta:
        model = WebsiteContact

    url = factory.Faker("uri")


class SocialMediaContactFactory(ContactBaseFactory):
    class Meta:
        model = SocialMediaContact

    platform = factory.Iterator(SocialMediaContact.Platforms)
    url = factory.Faker("url")



import markdown
from django.conf import settings
from markdown.extensions.tables import TableExtension
from markdown_blockdiag import BlockdiagExtension
from rest_framework.response import Response


def get_markdown_docs(docs):
    try:
        with open(docs, "r") as f:
            return Response(
                markdown.markdown(
                    f.read(),
                    extensions=settings.WBCORE_DEFAULT_MARKDOWN_EXTENSIONS,
                )
            )
    except FileNotFoundError:
        return Response(
            markdown.markdown(
                docs,
                extensions=settings.WBCORE_DEFAULT_MARKDOWN_EXTENSIONS,
            )
        )
    except AttributeError:
        return Response({"errors": "No documentation available."})



# Generated by Django 5.0.6 on 2024-07-10 11:34

from decimal import Decimal

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("example_app", "0003_change_stadium_capacity"),
    ]

    operations = [
        migrations.AlterField(
            model_name="player",
            name="transfer_value",
            field=models.DecimalField(
                decimal_places=2, default=Decimal("0"), max_digits=10, verbose_name="Market Value"
            ),
        ),
    ]



from enum import Enum
from typing import Union


class RequestType(Enum):
    GET = "get"
    POST = "post"
    DELETE = "delete"
    PATCH = "patch"
    PUT = "put"
    OPTIONS = "options"
    HEAD = "head"


class Button(Enum):
    # Buttons
    REFRESH = "refresh"
    NEW = "new"
    DELETE = "delete"

    # Buttons and Create Buttons
    SAVE = "save"
    SAVE_AND_CLOSE = "saveandclose"
    SAVE_AND_NEW = "saveandnew"

    # Create Buttons
    RESET = "reset"

    # Custom Buttons
    DROPDOWN = "dropdown"
    HYPERLINK = "hyperlink"
    WIDGET = "widget"
    ACTION = "action"

    @classmethod
    def buttons(cls):
        return [
            cls.REFRESH.value,
            cls.NEW.value,
            cls.DELETE.value,
            cls.SAVE.value,
            cls.SAVE_AND_CLOSE.value,
            cls.SAVE_AND_NEW.value,
        ]

    @classmethod
    def create_buttons(cls):
        return [
            cls.SAVE.value,
            cls.SAVE_AND_CLOSE.value,
            cls.SAVE_AND_NEW.value,
            cls.RESET.value,
        ]

    @classmethod
    def custom_buttons(cls):
        return [
            cls.DROPDOWN.value,
            cls.HYPERLINK.value,
            cls.WIDGET.value,
            cls.ACTION.value,
        ]


class WidgetType(Enum):
    LIST = "list"
    INSTANCE = "instance"
    CHART = "chart"
    SELECT = "select"


class Operator(Enum):
    EQUAL = "=="
    UNEQUAL = "!="
    GREATER = ">"
    GREATER_EQUAL = ">="
    LESS = "<"
    LESS_EQUAL = "<="
    EXISTS = "∃"


class Unit(Enum):
    FRACTION = "fr"
    REM = "rem"
    PIXEL = "px"

    def __call__(self, _value):
        return (float(_value), self.value)

    def unit(self, _value: Union[float, str, int]):
        assert isinstance(_value, (float, str, int)), "_value needs to be one of str, float or int"

        return f"{float(_value)}{self.value}"


class AuthType(Enum):
    NONE = "NONE"
    JWT = "JWT"



# Generated by Django 4.1.8 on 2023-04-20 13:09

import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("authentication", "0003_alter_user_profile"),
    ]

    operations = [
        migrations.CreateModel(
            name="Token",
            fields=[
                ("key", models.CharField(max_length=40, primary_key=True, serialize=False, verbose_name="Key")),
                ("created", models.DateTimeField(auto_now_add=True, verbose_name="Created")),
                ("valid_until", models.DateTimeField(blank=True, null=True, verbose_name="Valid Until")),
                (
                    "protected_view_name",
                    models.CharField(
                        blank=True,
                        help_text="If specified, only view that revert to this viewname will accept this token",
                        max_length=256,
                        null=True,
                        verbose_name="Protected View-Name",
                    ),
                ),
                (
                    "number_usage_left",
                    models.PositiveIntegerField(blank=True, null=True, verbose_name="Number of allowed usage"),
                ),
                ("is_valid", models.BooleanField(default=True, verbose_name="Is Valid")),
                (
                    "user",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="auth_tokens",
                        to=settings.AUTH_USER_MODEL,
                        verbose_name="User",
                    ),
                ),
            ],
            options={
                "verbose_name": "Token",
                "verbose_name_plural": "Tokens",
                "unique_together": {("user", "protected_view_name")},
            },
        ),
    ]



from django.utils.translation import gettext_lazy as _
from wbcore import filters
from wbcore.contrib.example_app.models import League, Team, TeamResults


class TeamResultsLeagueFilter(filters.FilterSet):
    team = filters.ModelChoiceFilter(
        label=_("Team"),
        queryset=Team.objects.all(),
        endpoint=Team.get_representation_endpoint(),
        value_key=Team.get_representation_value_key(),
        label_key=Team.get_representation_label_key(),
    )
    points__gte = filters.NumberFilter(label=_("Points"), lookup_expr="gte", field_name="points")
    points__lte = filters.NumberFilter(label=_("Points"), lookup_expr="lte", field_name="points")
    match_points_for__gte = filters.NumberFilter(
        label=_("Match Points For"), lookup_expr="gte", field_name="match_points_for"
    )
    match_points_for__lte = filters.NumberFilter(
        label=_("Match Points For"), lookup_expr="lte", field_name="match_points_for"
    )
    match_points_against__gte = filters.NumberFilter(
        label=_("Match Points Against"), lookup_expr="gte", field_name="match_points_against"
    )
    match_points_against__lte = filters.NumberFilter(
        label=_("Match Points Against"), lookup_expr="lte", field_name="match_points_against"
    )
    wins__gte = filters.NumberFilter(label=_("Wins"), lookup_expr="gte", field_name="wins")
    wins__lte = filters.NumberFilter(label=_("Points"), lookup_expr="lte", field_name="wins")
    draws__gte = filters.NumberFilter(label=_("Draws"), lookup_expr="gte", field_name="draws")
    draws__lte = filters.NumberFilter(label=_("Draws"), lookup_expr="lte", field_name="draws")
    losses__gte = filters.NumberFilter(label=_("Losses"), lookup_expr="gte", field_name="losses")
    losses__lte = filters.NumberFilter(label=_("Losses"), lookup_expr="lte", field_name="losses")

    class Meta:
        model = TeamResults
        fields = {}


class TeamResultsFilter(TeamResultsLeagueFilter):
    league = filters.ModelChoiceFilter(
        label=_("League"),
        queryset=League.objects.all(),
        endpoint=League.get_representation_endpoint(),
        value_key=League.get_representation_value_key(),
        label_key=League.get_representation_label_key(),
    )



from .condition import Condition
from .data import Data, DataValue
from .display import Display
from .process import Process, ProcessStep
from .step import (
    DecisionStep,
    EmailStep,
    FinishStep,
    JoinStep,
    ScriptStep,
    SplitStep,
    Step,
    UserStep,
    activate_step,
    StartStep,
    process_can_finish,
)
from .transition import Transition
from .workflow import Workflow



import pytest


@pytest.mark.django_db
class TestSpecificModels:
    pass

    # def test_get_by_natural_key(self, geography_factory):
    #
    #     obj = geography_factory(level=1)
    #     result = Geography.objects.get_by_natural_key(obj.code_2)
    #     assert obj == result



# Generated by Django 5.0.3 on 2024-04-24 11:28

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("notifications", "0004_alter_notification_body"),
    ]

    operations = [
        migrations.AlterField(
            model_name="notification",
            name="endpoint",
            field=models.CharField(blank=True, max_length=2048, null=True),
        ),
    ]



from rest_framework.request import Request
from rest_framework.reverse import reverse
from wbcore.configs.decorators import register_config


@register_config
def preset_config(request: Request) -> tuple[str, dict[str, str]]:
    """We return the wbcore:preset endpoint. We have to add a dummy arg and remove it immediately, since this endpoint only supports endpoints with arguments"""

    dummy_arg = "empty"
    arg_len = len(dummy_arg) + 1  # +1 to account for the second trailing slash
    endpoint = reverse("wbcore:preset", args=[dummy_arg], request=request)
    endpoint = endpoint[: -1 * arg_len]

    return "presets", {"endpoint": endpoint}



from django.urls import include, path
from wbcore.routers import WBCoreRouter

from . import viewsets

router = WBCoreRouter()

router.register(r"workflow", viewsets.WorkflowModelViewSet, basename="workflow")
router.register(r"workflowrepresentation", viewsets.WorkflowModelViewSet, basename="workflowrepresentation")
router.register(r"step", viewsets.StepModelViewSet, basename="step")
router.register(r"steprepresentation", viewsets.StepModelViewSet, basename="steprepresentation")
router.register(r"transition", viewsets.TransitionModelViewSet, basename="transition")
router.register(r"transitionrepresentation", viewsets.TransitionModelViewSet, basename="transitionrepresentation")
router.register(r"condition", viewsets.ConditionModelViewSet, basename="condition")
router.register(r"conditionrepresentation", viewsets.ConditionModelViewSet, basename="conditionrepresentation")
router.register(r"userstep", viewsets.UserStepModelViewSet, basename="userstep")
router.register(r"usersteprepresentation", viewsets.UserStepModelViewSet, basename="usersteprepresentation")
router.register(r"process", viewsets.ProcessModelViewSet, basename="process")
router.register(r"processrepresentation", viewsets.ProcessModelViewSet, basename="processrepresentation")
router.register(r"processstep", viewsets.ProcessStepModelViewSet, basename="processstep")
router.register(r"processsteprepresentation", viewsets.ProcessStepModelViewSet, basename="processsteprepresentation")
router.register(r"processstep-assigned", viewsets.AssignedProcessStepModelViewSet, basename="processstep-assigned")
router.register(r"display", viewsets.DisplayModelViewSet, basename="display")
router.register(r"displayrepresentation", viewsets.DisplayModelViewSet, basename="displayrepresentation")
router.register(r"decisionstep", viewsets.DecisionStepModelViewSet, basename="decisionstep")
router.register(
    r"decisionsteprepresentation", viewsets.DecisionStepModelViewSet, basename="decisionsteprepresentation"
)
router.register(r"startstep", viewsets.StartStepModelViewSet, basename="startstep")
router.register(r"startsteprepresentation", viewsets.StartStepModelViewSet, basename="startsteprepresentation")
router.register(r"splitstep", viewsets.SplitStepModelViewSet, basename="splitstep")
router.register(r"splitsteprepresentation", viewsets.SplitStepModelViewSet, basename="splitsteprepresentation")
router.register(r"joinstep", viewsets.JoinStepModelViewSet, basename="joinstep")
router.register(r"joinsteprepresentation", viewsets.JoinStepModelViewSet, basename="joinsteprepresentation")
router.register(r"scriptstep", viewsets.ScriptStepModelViewSet, basename="scriptstep")
router.register(r"scriptsteprepresentation", viewsets.ScriptStepModelViewSet, basename="scriptsteprepresentation")
router.register(r"emailstep", viewsets.EmailStepModelViewSet, basename="emailstep")
router.register(r"emailsteprepresentation", viewsets.EmailStepModelViewSet, basename="emailsteprepresentation")
router.register(r"finishstep", viewsets.FinishStepModelViewSet, basename="finishstep")
router.register(r"finishsteprepresentation", viewsets.FinishStepModelViewSet, basename="finishsteprepresentation")
router.register(r"data", viewsets.DataModelViewSet, basename="data")
router.register(r"datarepresentation", viewsets.DataModelViewSet, basename="datarepresentation")

step_router = WBCoreRouter()
step_router.register(r"transition-step", viewsets.TransitionModelViewSet, basename="transition-step")
step_router.register(r"processstep-step", viewsets.ProcessStepModelViewSet, basename="processstep-step")

workflow_router = WBCoreRouter()
workflow_router.register(r"transition-workflow", viewsets.TransitionModelViewSet, basename="transition-workflow")
workflow_router.register(r"step-workflow", viewsets.StepModelViewSet, basename="step-workflow")
workflow_router.register(r"process-workflow", viewsets.ProcessModelViewSet, basename="process-workflow")
workflow_router.register(r"data-workflow", viewsets.DataModelViewSet, basename="data-workflow")

transition_router = WBCoreRouter()
transition_router.register(r"condition-transition", viewsets.ConditionModelViewSet, basename="condition-transition")

process_router = WBCoreRouter()
process_router.register(r"processstep-process", viewsets.ProcessStepModelViewSet, basename="processstep-process")

urlpatterns = [
    path("", include(router.urls)),
    path("step/<int:step_id>/", include(step_router.urls)),
    path("workflow/<int:workflow_id>/", include(workflow_router.urls)),
    path("transition/<int:transition_id>/", include(transition_router.urls)),
    path("process/<str:process_id>/", include(process_router.urls)),
]



from django.db.models import CharField


class IconField(CharField):
    description = "A field storing an icon"

    def deconstruct(self):
        name, path, args, kwargs = super().deconstruct()
        if "choices" in kwargs:
            del kwargs["choices"]
        return name, path, args, kwargs



from django.utils.translation import gettext as _
from wbcore.metadata.configs.titles import TitleViewConfig


class RoleTitleConfig(TitleViewConfig):
    def get_list_title(self) -> str:
        return _("Roles")

    def get_create_title(self) -> str:
        return _("Create Role")

    def get_instance_title(self) -> str:
        return _("Role")



from django.db.models import Count
from wbcore import viewsets
from wbcore.contrib.documents.filters import DocumentTypeFilter
from wbcore.contrib.documents.models import DocumentType
from wbcore.contrib.documents.serializers import (
    DocumentTypeModelSerializer,
    DocumentTypeRepresentationSerializer,
)
from wbcore.contrib.documents.viewsets.buttons import DocumentTypeButtonConfig
from wbcore.contrib.documents.viewsets.display import DocumentTypeModelDisplay
from wbcore.contrib.documents.viewsets.titles import DocumentTypeModelTitleConfig


class DocumentTypeModelViewSet(viewsets.ModelViewSet):
    LIST_DOCUMENTATION = "documents/markdown/documentation/document_types.md"
    search_fields = ("name",)
    ordering = ["name"]
    ordering_fields = (
        "name",
        "parent",
    )
    filterset_class = DocumentTypeFilter
    queryset = DocumentType.objects.all()
    serializer_class = DocumentTypeModelSerializer
    title_config_class = DocumentTypeModelTitleConfig

    button_config_class = DocumentTypeButtonConfig
    display_config_class = DocumentTypeModelDisplay

    def get_queryset(self):
        return super().get_queryset().annotate(document_count=Count("documents"))


class DocumentTypeRepresentationViewSet(viewsets.RepresentationViewSet):
    queryset = DocumentType.objects.all()
    serializer_class = DocumentTypeRepresentationSerializer
    search_fields = ("name",)



from django.urls import include, path
from wbcore.routers import WBCoreRouter

from . import viewsets

router = WBCoreRouter()

router.register(r"document", viewsets.DocumentModelViewSet, basename="document")
router.register(r"documenttype", viewsets.DocumentTypeModelViewSet, basename="documenttype")
router.register(r"link", viewsets.ShareableLinkModelViewSet, basename="link")
router.register(r"linkaccess", viewsets.ShareableLinkAccessModelViewSet, basename="linkaccess")
router.register(
    r"documentmodelrelationship", viewsets.DocumentModelRelationshipModelViewSet, basename="documentmodelrelationship"
)

# Representations
router.register(r"documentrepresentation", viewsets.DocumentRepresentationViewSet, basename="documentrepresentation")
router.register(
    r"documenttyperepresentation", viewsets.DocumentTypeRepresentationViewSet, basename="documenttyperepresentation"
)
router.register(r"linkrepresentation", viewsets.ShareableLinkRepresentationViewSet, basename="linkrepresentation")
router.register(
    r"linkaccessrepresentation", viewsets.ShareableLinkAccessRepresentationViewSet, basename="linkaccessrepresentation"
)

document_router = WBCoreRouter()
document_router.register(
    r"documentmodelrelationship",
    viewsets.DocumentModelRelationshipModelViewSet,
    basename="document-documentmodelrelationship",
)

urlpatterns = [
    path("", include(router.urls)),
    path("document/<int:document_id>/", include(document_router.urls)),
    path("download/<uuid:uuid>/", viewsets.DownloadShareableLinkView.as_view(), name="download"),
    path(
        "contentdocument/<int:content_type>/<str:content_id>/",
        viewsets.DocumentModelViewSet.as_view({"get": "list", "post": "create"}),
        name="document_content_object",
    ),
]



from django.urls import include, path
from wbcore.routers import WBCoreRouter

from .viewsets import (
    CurrencyFXRatesCurrencyModelViewSet,
    CurrencyModelViewSet,
    CurrencyRepresentationViewSet,
)

router = WBCoreRouter()
router.register(r"currency", CurrencyModelViewSet, basename="currency")
router.register(r"currencyrepresentation", CurrencyRepresentationViewSet, basename="currencyrepresentation")


currency_router = WBCoreRouter()
currency_router.register(r"currencyfxrates", CurrencyFXRatesCurrencyModelViewSet, basename="currency-currencyfxrates")


urlpatterns = [path("", include(router.urls)), path("currency/<int:currency_id>/", include(currency_router.urls))]



from unittest.mock import patch

import pytest
from anymail.inbound import AnymailInboundMessage
from anymail.signals import AnymailInboundEvent, EventType, inbound
from django.utils import timezone
from faker import Faker

from ..backends.mail import DataBackend as MailDataBackend
from ..backends.sftp import DataBackend as SFTPDataBackend
from ..models import Source
from .test_models import get_byte_stream

fake = Faker()


def conn(sftpsrv):
    """return a dictionary holding argument info for the sftp client"""
    return {
        "host": sftpsrv.host,
        "port": sftpsrv.port,
        "username": "user",
        "password": "pw",
        "default_path": "/home/test",
    }


# filesystem served by pytest-sftpserver plugin
VFS = {
    "home": {
        "test": {
            "pub": {
                "make.txt": "content of make.txt",
                "foo1.csv": "CSV Content",
                "image01.jpg": "data for image01.jpg",
            },
            "read.me": "contents of read.me",
        }
    }
}


@pytest.mark.django_db
class TestBackend:
    def test_mail_backend(self):
        b1 = get_byte_stream()
        b2 = get_byte_stream()
        filename1 = "file_valid_12385.json"
        filename2 = "file.json"
        att1 = AnymailInboundMessage.construct_attachment(
            "application/json", b1.getvalue(), charset="utf-8", filename=filename1
        )

        att2 = AnymailInboundMessage.construct_attachment(
            "application/json", b2.getvalue(), charset="utf-8", filename=filename2
        )

        msg = AnymailInboundMessage.construct(
            from_email="from@example.com",
            to="to@example.com",
            cc="cc@example.com",
            subject="test subject",
            attachments=[att1, att2],
        )
        backend = MailDataBackend()
        res = list(backend.get_files(timezone.now(), file_name_regex="file_valid_([A-Z0-9]*).json", message=msg))
        assert len(res) == 1
        assert res[0][0] == filename1
        assert (res[0][1]).getvalue() == b1.getvalue()

    @pytest.mark.parametrize(
        "source__import_parameters,data_backend__backend_class_path,from_email",
        [
            ({"whitelisted_emails": ["test@test.ch"]}, "wbcore.contrib.io.backends.mail", "test@test.ch"),
            ({"whitelisted_emails": ["test@test.ch"]}, "wbcore.contrib.io.backends.mail", "spam@test.ch"),
            ({"whitelisted_emails": ["test@test.ch"]}, "wbcore.contrib.io.backends.unvalid", "test@test.ch"),
        ],
    )
    @patch.object(Source, "generate_import_sources")
    def test_inbound_mail(self, mock_process_source, source, data_backend, from_email):
        source.data_backend = data_backend
        source.save()
        message = AnymailInboundMessage.construct(
            from_email=from_email,
            to="to@example.com",
            subject=fake.paragraph(),
            text=fake.paragraph(),
            html=f"<p>{fake.paragraph()}</p>",
        )
        event = AnymailInboundEvent(
            event_type=EventType.INBOUND,
            timestamp=None,
            event_id=fake.uuid4(),
            esp_event={"from_email": from_email},
            message=message,
        )
        inbound.send(sender="TEST", event=event, esp_name="test")
        if (
            source.data_backend.backend_class_path == "wbcore.contrib.io.backends.mail"
            and source.import_parameters.get("whitelisted_emails", [None])[0] == from_email
        ):
            assert mock_process_source.call_count == 1
        else:
            assert mock_process_source.call_count == 0

    def test_sftp_backend_without_credential(self):
        with pytest.raises(ValueError):
            SFTPDataBackend()

    def test_sftp_backend(self, sftpserver, import_credential_factory):
        with sftpserver.serve_content(VFS):
            import_credential = import_credential_factory.create(
                username="user",
                password="password",
                additional_resources={
                    "host": sftpserver.host,
                    "port": sftpserver.port,
                },
            )
            copied_file_content = VFS["home"]["test"]["pub"]["foo1.csv"]
            backend = SFTPDataBackend(import_credential=import_credential)
            res = list(
                backend.get_files(timezone.now(), "/home/test/pub", file_name_regex=r".+\.csv$", cleanup_files=True)
            )

            assert len(res) == 1
            assert res[0][0] == "foo1.csv"
            assert (res[0][1]).getvalue().decode("utf-8") == copied_file_content



from django.utils.translation import gettext as _
from wbcore.contrib.directory.models import BankingContact
from wbcore.menus import ItemPermission, MenuItem
from wbcore.permissions.shortcuts import is_internal_user

PENDING_BANKINGCONTACT_MENUITEM = MenuItem(
    label=_("Pending Banking Contacts"),
    endpoint="wbcore:directory:bankingcontact-list",
    endpoint_get_parameters={"status": BankingContact.Status.PENDING.value},
    permission=ItemPermission(
        method=lambda request: is_internal_user(request.user),
        permissions=[
            "directory.view_bankingcontact",
            "directory.administrate_banking_contact",
        ],
    ),
)

TELEPHONECONTACTSEARCH_MENUITEM = MenuItem(
    label=_("Search by Telephone Contact"),
    endpoint="wbcore:directory:telephonecontact-list",
    permission=ItemPermission(
        method=lambda request: is_internal_user(request.user), permissions=["directory.view_telephonecontact"]
    ),
)



import datetime
import json
import os
from contextlib import suppress
from datetime import timedelta
from typing import Dict, Generator, List, Optional

import magic
import reversion
from celery import shared_task
from django.conf import settings
from django.contrib.contenttypes.models import ContentType
from django.core.mail import EmailMultiAlternatives
from django.db import models
from django.template import Context, Template
from django.template.loader import get_template
from django.utils import timezone
from django.utils.functional import cached_property
from django.utils.html import strip_tags
from django.utils.translation import gettext
from django.utils.translation import gettext_lazy as _
from slugify import slugify
from wbcore.contrib.authentication.models import User
from wbcore.contrib.guardian.models.mixins import PermissionObjectModelMixin
from wbcore.models import WBModel
from wbcore.utils.html import convert_html2text

from .document_model_relationships import DocumentModelRelationship
from .shareable_links import ShareableLink


def upload_to(instance: "Document", filename: str) -> str:
    """
    Return the document path based on the document type mptt tree
    Args:
        instance: A document instance
        filename: The given filename

    Returns:
        The relative document path
    """

    if instance.document_type:
        type_ancestors = instance.document_type.get_ancestors(ascending=False, include_self=True).values_list(
            "name", flat=True
        )
        base_path = "/".join(["dms/document", *[slugify(ancestor_name) for ancestor_name in type_ancestors]])
    else:
        base_path = "dms/document"
    return f"{base_path}/{filename}"


@reversion.register()
class Document(PermissionObjectModelMixin, WBModel):
    name = models.CharField(max_length=255)
    description = models.TextField(default="", blank=True, verbose_name=_("Description"))
    file = models.FileField(max_length=256, upload_to=upload_to, verbose_name=_("File"))
    document_type = models.ForeignKey(
        to="documents.DocumentType",
        related_name="documents",
        on_delete=models.PROTECT,
        verbose_name=_("Document Type"),
    )
    system_created = models.BooleanField(default=False, verbose_name=_("System Created"))
    system_key = models.CharField(max_length=255, null=True, blank=True, verbose_name=_("System Key"), unique=True)

    valid_from = models.DateField(null=True, blank=True, verbose_name=_("Valid From"))
    valid_until = models.DateField(null=True, blank=True, verbose_name=_("Valid Until"))

    created = models.DateTimeField(auto_now_add=True, verbose_name=_("Created"))
    updated = models.DateTimeField(auto_now=True, blank=True, null=True, verbose_name=_("Updated"))

    def __str__(self) -> str:
        return f"{self.name} ({self.document_type})"

    @cached_property
    def content_type(self) -> str:
        """
        Guess the file content type using magic library

        Returns:
            The file content type

        Raises:
            MagicException: If content type cannot be guessed
        """
        try:
            return magic.from_buffer(self.file.read(), mime=True)
        except magic.MagicException:
            return None

    @cached_property
    def filename(self) -> str:
        """
        Return the extracted filename from the relative path

        Returns:
            The filename
        """
        return os.path.basename(self.file.name)

    @property
    def linked_objects(self) -> Generator[models.Model, None, None]:
        """
        All objects that share a relationship with this document instance

        Returns:
            An generator of Object (any type)

        """
        for relation in self.relationships.all():
            with suppress(
                AttributeError
            ):  # maybe the content type doesn't exist anymore and getting content_object will then trigger an attribute error that we want to silently catch
                yield relation.content_object

    def get_permissions_for_user(self, user: "User", created: Optional[datetime.datetime] = None) -> Dict[str, bool]:
        """
        Return a generator of allowed (view|change|delete) permission and its editable state

        Args:
            user: The user to which we get permission for the given object
            created: The permission creation date. Defaults to None.

        Returns:
            A dictionary of string permission identifier, editable state key value pairs.
        """
        # We do not call super here, as we do not want that by default everyone has access to all documents
        base_permission = {}
        if not self.system_created:
            base_permission = super().get_permissions_for_user(user, created=created)

        for linked_object in self.linked_objects:
            if _callable := getattr(linked_object, "get_permissions_for_user_and_document", None):
                extra_permissions = _callable(user, self, created=created)
                for perm, editable in extra_permissions:
                    base_permission[perm] = editable
        return base_permission

    def generate_shareable_link(
        self, sharing_seconds_duration: Optional[int] = 3600, one_time_link: Optional[bool] = False
    ) -> "ShareableLink":
        """
        Generate a shareable link instance.

        Args:
            sharing_seconds_duration: The validity in second. Convert this value into a valid datetime. Defaults to 1 hour.
            one_time_link: Set to True if this link can be accessed only once. Defaults to False.

        Returns:
            The newly created shareable link instance
        """
        return ShareableLink.objects.create(
            document=self,
            valid_until=(
                timezone.now() + timedelta(seconds=sharing_seconds_duration) if sharing_seconds_duration else None
            ),
            one_time_link=one_time_link,
        )

    def link(self, obj: models.Model) -> bool:
        """
        Link a document to the passed object

        Args:
            obj: The object that is related to the document

        Returns:
            True if the relationship was created. False otherwise.
        """

        return DocumentModelRelationship.objects.update_or_create(
            document=self,
            content_type=ContentType.objects.get_for_model(obj),
            object_id=obj.id,
            defaults={"primary": True},
        )[0]

    def send_email(
        self,
        to_emails: List | str,
        as_link: Optional[bool] = True,
        subject: Optional[str] = None,
        from_email: Optional[str] = None,
        body: Optional[str] = None,
        cc_emails: Optional[List[str]] = None,
        bcc_emails: Optional[List[str]] = None,
    ):
        """
        Send a document through the configured mailing backend

        Args:
            to_emails: A list of emails or an email address
            as_link: True if the document needs to be sent with a shareable link (Otherwise attached as file). Defaults to True.
            subject: The email subject. Defaults to None.
            from_email: The sender email address. Defaults to None.
            body: Mail Content: Defaults to None.
            cc_emails: CC addresses. Defaults to None.
            bcc_emails: BCC addresses. Defaults to None.
        """
        send_email_as_task.delay(self.id, to_emails, as_link, subject, from_email, body, cc_emails, bcc_emails)

    @classmethod
    def get_for_object(cls, obj) -> "models.QuerySet[Document]":
        """
        Returns a Queryset of document linked to the passed object

        Args:
            obj: The related object

        Returns:
            A queryset of documents
        """
        document_ids = DocumentModelRelationship.objects.filter(
            content_type=ContentType.objects.get_for_model(obj), object_id=obj.id
        ).values("document")
        return cls.objects.filter(id__in=document_ids)

    @classmethod
    def get_endpoint_basename(cls):
        return "wbcore:documents:document"

    @classmethod
    def get_representation_value_key(cls):
        return "id"

    @classmethod
    def get_representation_endpoint(cls):
        return "wbcore:documents:documentrepresentation-list"

    @classmethod
    def get_representation_label_key(cls):
        return "{{name}}"

    class Meta(PermissionObjectModelMixin.Meta):
        verbose_name = _("Document")
        verbose_name_plural = _("Documents")


@shared_task
def send_email_as_task(
    document_id,
    to_emails: List | str,
    as_link: Optional[bool] = True,
    subject: Optional[str] = None,
    from_email: Optional[str] = None,
    body: Optional[str] = None,
    cc_emails: Optional[List[str]] = None,
    bcc_emails: Optional[List[str]] = None,
):
    """
    Send a document through the configured mailing backend

    Args:
        document_id: The ID of the document.
        to_emails: A list of emails or an email address
        as_link: True if the document needs to be sent with a shareable link (Otherwise attached as file). Defaults to True.
        subject: The email subject. Defaults to the document name.
        from_email: The sender email address. Defaults to the global preferences
        body: The mail content. Defaults to the default template.
        cc_emails:  CC addresses. Defaults to None.
        bcc_emails: BCC addresses. Defaults to None.
    """
    document = Document.objects.get(id=document_id)
    if not from_email:
        from_email = settings.DEFAULT_FROM_EMAIL
    if not subject:
        subject = document.name
    context = {"title": gettext("A file was shared with you")}
    if as_link:
        link = document.generate_shareable_link(sharing_seconds_duration=0).link
        context["endpoint"] = link
        context["message"] = gettext(
            """
        <p>Please find your {document_type} \"{filename}\" under this <a href={link}>link</a></p>
        """
        ).format(document_type=document.document_type, filename=document.filename, link=link)
    else:
        context["message"] = gettext(
            """
        <p>Please find the {document_type} \"{filename}\" attached to this mail.</p>
        """
        ).format(document_type=document.document_type, filename=document.filename)
    if body:
        html_content = Template(body).render(Context(context))
    else:
        template = get_template(settings.WBCORE_NOTIFICATION_TEMPLATE)
        html_content = template.render(context)

    if not isinstance(to_emails, list):
        to_emails = [to_emails]

    msg = EmailMultiAlternatives(
        strip_tags(subject),
        body=convert_html2text(html_content),
        to=to_emails,
        from_email=from_email,
        cc=cc_emails,
        bcc=bcc_emails,
    )
    msg.attach_alternative(html_content, "text/html")

    if not as_link:
        alternative_dict = {"id": document.id, "filename": document.filename}
        msg.attach_alternative(
            json.dumps(alternative_dict).encode("ascii"), "wbdms/document"
        )  # leave this mimetype. Used by wbmailing
        msg.attach(document.filename, document.file.read())
    msg.send()



from contextlib import suppress

from celery import shared_task
from django.apps import apps
from django.contrib.auth.backends import ModelBackend
from django.contrib.postgres.fields import DateTimeRangeField
from django.db import models, transaction
from django.dispatch import receiver
from django.utils.translation import gettext_lazy as _
from ics import Event
from wbcore.contrib.color.fields import ColorField
from wbcore.contrib.icons import WBIcon
from wbcore.contrib.icons.models import IconField
from wbcore.models import WBModel
from wbcore.utils.itertools import get_inheriting_subclasses
from wbcore.utils.models import DeleteToDisableMixin


class CalendarItem(DeleteToDisableMixin, WBModel):
    AUTOMATICALLY_CLEAN_SOFT_DELETED_OBJECTS = True

    class Visibility(models.TextChoices):
        PUBLIC = "PUBLIC", _("Public")
        PRIVATE = "PRIVATE", _("Private")
        CONFIDENTIAL = "CONFIDENTIAL", _("Confidential")

    color = ColorField(
        blank=True,
        null=True,
    )
    entities = models.ManyToManyField(
        to="directory.Entry", verbose_name=_("Entities"), related_name="calendar_entities"
    )
    entity_list = models.JSONField(blank=True, null=True)
    icon = IconField(
        max_length=128,
        verbose_name=_("Icon"),
        blank=True,
        null=True,
    )
    visibility = models.CharField(
        max_length=255,
        choices=Visibility.choices,
        verbose_name=_("Visibility"),
        default=Visibility.PUBLIC,
    )
    item_type = models.CharField(max_length=255, verbose_name=_("Type"), blank=True, null=True)

    period = DateTimeRangeField(
        blank=True,
        null=True,
        # Translator: In this context, it is a period of time
        verbose_name=_("Period"),
    )
    title = models.CharField(max_length=255, verbose_name=_("Title"))
    conference_room = models.ForeignKey(
        to="agenda.ConferenceRoom",
        verbose_name=_("Conference Room"),
        related_name="calendar_item",
        on_delete=models.SET_NULL,
        null=True,
        blank=True,
    )
    all_day = models.BooleanField(
        default=False, verbose_name=_("All Day"), help_text=_("Check this if the activity spans throughout the day")
    )
    is_cancelled = models.BooleanField(default=False, verbose_name=_("Is Cancelled"))
    is_deletable = models.BooleanField(default=True, verbose_name=_("Is Deletable"))

    def get_casted_calendar_item(self) -> models.Model:
        """
        Cast the calendar item into its child representative

        Raises ValueError, AttributeError or LookupError on wrong item_type
        """
        model = apps.get_model(self.item_type)
        return model.objects.get(pk=self.pk)

    def get_color(self) -> str:
        """Sets a default color value. This method should be overridden by subclasses.

        Returns:
            str: black as hex string
        """

        return "#000000"

    def get_icon(self) -> str:
        """Sets a default icon value. This method should be overridden by subclasses

        Returns:
            str: event icon string based on the icon backend
        """

        return WBIcon.EVENT.icon

    @classmethod
    def get_item_types_choices(cls) -> tuple[str, str]:
        """
        Iterate over all implementing classes and extract its item type
        """
        choices = []
        for snd in get_inheriting_subclasses(CalendarItem):
            if (
                (_meta := getattr(snd, "_meta", None))
                and hasattr(snd, "get_item_type")
                and callable(snd.get_item_type)
            ):
                choices.append((snd.get_item_type(), _meta.verbose_name))
        return choices

    @classmethod
    def get_item_type(cls) -> str:
        # Problems with the app label can occur with nested apps. To prevent these cases,
        # we must ensure that only the last part of the app label is used.
        app_label: str = cls._meta.app_label.split(".")[-1]
        return f"{app_label}.{cls.__name__}"

    def __str__(self):
        return self.title

    def save(self, *args, **kwargs):
        if not self.item_type:
            self.item_type = self.get_item_type()

        self.color = self.get_color()
        self.icon = self.get_icon()
        super().save(*args, **kwargs)

    @property
    def endpoint_basename(self) -> str | None:
        with suppress(ValueError, AttributeError, LookupError):
            return apps.get_model(self.item_type).get_endpoint_basename()

    def to_ics(self, start=None, end=None):
        """
        Convert calendar item model to ICS event

        Arguments:
            start {datetime.date} -- Start calendar item time if self.start is None
            end {datetime.date} -- End calendar item time if self.end is None
        Returns:
            Event -- The calendar item as ICS event
        """
        e = Event()
        if self.period:
            if not start:
                start = self.period.lower  # type: ignore
            if not end:
                end = self.period.upper  # type: ignore
        if (start and end) and (start < end):
            e.name = self.title
            e.begin = start
            e.end = end

            _id = f"workbench-{self.id}-{start}-{end}"
            e.uid = _id.replace(" ", "_")
            with suppress(ValueError, AttributeError, LookupError):
                casted_calendar_item = self.get_casted_calendar_item()
                if hasattr(casted_calendar_item, "get_extra_ics_kwargs") and callable(
                    casted_calendar_item.get_extra_ics_kwargs
                ):
                    for k, v in casted_calendar_item.get_extra_ics_kwargs():
                        setattr(e, k, v)

                return e
        else:
            return None

    def can_delete(self, user):
        try:
            return self.get_casted_calendar_item().is_deletable and (
                self.entities.filter(id=user.profile.id).exists()
                or self.has_user_administrate_permission(user, include_superusers=True)
            )
        except (ValueError, AttributeError, LookupError):
            return False

    @classmethod
    def get_endpoint_basename(cls):
        return "wbcore:agenda:calendaritem"

    @classmethod
    def get_representation_endpoint(cls):
        return "wbcore:agenda:calendaritemrepresentation-list"

    @classmethod
    def get_representation_value_key(cls):
        return "id"

    @classmethod
    def get_representation_label_key(cls):
        return "{{title}}"

    @classmethod
    def has_user_administrate_permission(cls, user, include_superusers: bool = False) -> bool:
        return user in ModelBackend().with_perm(
            "agenda.administrate_confidential_items", include_superusers=include_superusers
        )

    @property
    def duration(self):
        if self.period.lower and self.period.upper:
            return self.period.upper - self.period.lower

    class Meta:
        verbose_name = _("Calendar Item")
        verbose_name_plural = _("Calendar Items")
        permissions = [
            ("administrate_confidential_items", "Can administrate calendar items"),
        ]
        indexes = [models.Index(fields=["period"])]


@receiver(models.signals.m2m_changed, sender=CalendarItem.entities.through)
def m2m_entities_changed(sender, instance, action, pk_set, **kwargs):
    if action in ["post_add", "post_remove", "post_clear"]:
        transaction.on_commit(set_entity_list.s(instance.id).delay)


@shared_task
def set_entity_list(calendar_item_id: int):
    frontend_list = []
    with suppress(
        CalendarItem.DoesNotExist
    ):  # for sanitization reason, we catch this error in case of db slowness. It doesn't really matter as this is computed frequently.
        instance = CalendarItem.objects.get(id=calendar_item_id)
        for entity in instance.entities.exclude(entry_type="Company"):
            person = entity.get_casted_entry()
            frontend_list.append(
                {
                    "id": person.id,
                    "label": person.initials,
                    "tooltip": person.computed_str,
                }
            )
        instance.entity_list = frontend_list
        instance.save()



from datetime import date, datetime, timedelta

import pytest
from django.forms.models import model_to_dict
from rest_framework.exceptions import ValidationError
from rest_framework.test import APITestCase
from wbcore.contrib.directory.factories import PersonFactory
from wbcore.contrib.example_app.factories import (
    PlayerFactory,
    SportPersonFactory,
    StadiumFactory,
    TeamFactory,
)
from wbcore.contrib.example_app.serializers import (
    PlayerModelSerializer,
    SportPersonModelSerializer,
    TeamErrorMessages,
    TeamModelSerializer,
)


@pytest.mark.django_db
class TestSportPersonModelSerializer(APITestCase):
    def test_person_serializer(self):
        person_data: dict = model_to_dict(SportPersonFactory.build(profile=PersonFactory()))
        sport_person_serializer = SportPersonModelSerializer(data=person_data)
        self.assertTrue(sport_person_serializer.is_valid())

    def test_player_serializer(self):
        player_data: dict = model_to_dict(PlayerFactory.build(current_team=TeamFactory()))
        player_serializer = PlayerModelSerializer(data=player_data)
        self.assertTrue(player_serializer.is_valid())


@pytest.mark.django_db
class TestTeamModelSerializer(APITestCase):
    def test_team_serializer(self):
        coach = SportPersonFactory()
        stadium = StadiumFactory()

        team_data: dict = model_to_dict(TeamFactory.build(coach=coach, home_stadium=stadium))
        team_serializer = TeamModelSerializer(data=team_data)
        self.assertTrue(team_serializer.is_valid())

    def test_team_exists(self):
        coach = SportPersonFactory()
        stadium = StadiumFactory()
        team = TeamFactory(coach=coach, home_stadium=stadium)
        new_team_data: dict = model_to_dict(TeamFactory.build(home_stadium=stadium, name=team.name))
        with self.assertRaisesMessage(ValidationError, TeamErrorMessages.team_exists.value):
            team_serializer = TeamModelSerializer(data=new_team_data)
            self.assertFalse(team_serializer.is_valid(raise_exception=True))

    def test_team_wrong_date(self):
        stadium = StadiumFactory()
        future_date: date = (datetime.now() + timedelta(days=1)).date()
        team_data: dict = model_to_dict(TeamFactory.build(founded_date=future_date, home_stadium=stadium))

        with self.assertRaisesMessage(ValidationError, TeamErrorMessages.wrong_founding_date.value):
            team_serializer = TeamModelSerializer(data=team_data)
            self.assertFalse(team_serializer.is_valid(raise_exception=True))

    def test_team_name_placeholder(self):
        coach = SportPersonFactory()
        stadium = StadiumFactory()
        team_data: dict = model_to_dict(TeamFactory.build(home_stadium=stadium, coach=coach))
        team_serializer = TeamModelSerializer(data=team_data)
        self.assertTrue(team_serializer.is_valid())
        self.assertTrue("name" in team_serializer.fields)
        self.assertTrue(team_serializer.fields["name"].placeholder == "Enter team name here")



import urllib
from datetime import datetime, timedelta

from django.conf import settings
from django.db import models
from django.urls import resolve
from django.utils import timezone
from django.utils.translation import gettext_lazy as _
from dynamic_preferences.models import global_preferences_registry
from rest_framework import exceptions
from rest_framework.authtoken.models import Token as DRFToken

from .users import User


class Token(DRFToken):
    """
    Short-Lived and short-usage Token to be used as DRF token
    """

    valid_until = models.DateTimeField(blank=True, null=True, verbose_name=_("Valid Until"))
    protected_view_name = models.CharField(
        blank=True,
        null=True,
        max_length=256,
        verbose_name="Protected View-Name",
        help_text="If specified, only view that revert to this viewname will accept this token",
    )
    number_usage_left = models.PositiveIntegerField(blank=True, null=True, verbose_name=_("Number of allowed usage"))
    is_valid = models.BooleanField(default=True, verbose_name=_("Is Valid"))
    user = models.ForeignKey(
        settings.AUTH_USER_MODEL, related_name="auth_tokens", on_delete=models.CASCADE, verbose_name=_("User")
    )

    def __str__(self):
        if self.key:
            return self.key
        return ""

    def check_validity_for_endpoint(self, view_name: str, current_time: datetime) -> True:
        """
        Handle the logic for token validity. Expect this method to be called from a request access
        Args:
            view_name: view name to check authentication against
            current_time: The time to check authentication validity

        Returns:
            True if the token is valid and authentication can be granted
        """
        if not current_time:
            current_time = timezone.now()
        if not self.is_valid:
            raise exceptions.AuthenticationFailed("Token is not valid anymore")
        if self.valid_until is not None and self.valid_until < current_time:
            raise exceptions.AuthenticationFailed("Token has expired")

        if self.protected_view_name and not view_name == self.protected_view_name:
            raise exceptions.AuthenticationFailed("Token has limited scope that doesn't match the requested endpoint")
        if self.number_usage_left is not None:
            if self.number_usage_left <= 0:
                raise exceptions.AuthenticationFailed("Token usage exhausted")
            self.number_usage_left -= 1
            self.save()
        return True

    @classmethod
    def generate_shareable_link(
        cls,
        endpoint: str,
        user: "User",
        creation_time: datetime | None = None,
        protected_view_name: str = None,
        token: str | None = None,
    ) -> str:
        """
        Given an endpoint and a user, generate a shorted live token valid from now to the default validity time and appends it to the base endpoint query parameters
        Args:
            endpoint: endpoint to append the token to
            user: token's owner
        Returns:
            an valid endpoint with the token as query parameter
        """
        if not creation_time:
            creation_time = timezone.now()
        global_preferences = global_preferences_registry.manager()
        valid_until = creation_time + timedelta(
            hours=global_preferences["authentication__default_token_validity_in_hours"]
        )

        parse_url = urllib.parse.urlsplit(urllib.parse.unquote(endpoint))
        endpoint = parse_url.path
        base_params = dict(urllib.parse.parse_qsl(parse_url.query))
        if not token:
            token = Token.generate_token_for_user(
                user,
                valid_until=valid_until,
                protected_view_name=resolve(endpoint).view_name if not protected_view_name else protected_view_name,
            ).key
        base_params["token"] = token

        if parse_url.scheme and parse_url.netloc:
            endpoint = parse_url.scheme + "://" + parse_url.netloc + endpoint
        params_repr = "&".join([f"{k}={v}" for k, v in base_params.items()])
        return endpoint + f"?{params_repr}"

    @classmethod
    def generate_token_for_user(
        cls,
        user: "User",
        valid_until: datetime | None = None,
        number_of_usages: int | None = None,
        protected_view_name: str | None = None,
    ):
        """
        Generate a token for a user. We assume uniqueness between user and protected_view_name to avoid crowding our table.
        Args:
            user: Token's owner
            valid_until: Token maximum datetime validity
            number_of_usages: Token maximum token usages
            protected_view_name: Token restricted view
        Returns:
            A valid Token
        """

        token, created = Token.objects.get_or_create(
            user=user,
            protected_view_name=protected_view_name,
            defaults={"valid_until": valid_until, "number_usage_left": number_of_usages},
        )
        if not created:
            token.number_usage_left = number_of_usages
            token.valid_until = valid_until
            token.is_valid = True
            token.save()
        return token

    class Meta:
        verbose_name = _("Token")
        verbose_name_plural = _("Tokens")
        unique_together = [("user", "protected_view_name")]



# Generated by Django 5.0.5 on 2024-06-11 10:35

import django.db.models.deletion
from django.db import migrations, models


def add_missing_persons(apps, schema_editor):
    Person = apps.get_model("directory", "Person")
    SportPerson = apps.get_model("example_app", "SportPerson")

    for sport_person in SportPerson.objects.filter(profile__isnull=True):
        person = Person.objects.create(
            first_name=sport_person.first_name,
            last_name=sport_person.last_name,
        )
        sport_person.profile = person
        sport_person.save()


class Migration(migrations.Migration):
    dependencies = [
        ("directory", "0007_alter_bankingcontact_options"),
        ("example_app", "0001_initial"),
    ]

    operations = [
        migrations.AddField(
            model_name="sportperson",
            name="profile",
            field=models.OneToOneField(
                blank=True,
                default=None,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="sport_profile",
                to="directory.person",
            ),
        ),
        migrations.RunSQL(sql="SET CONSTRAINTS ALL IMMEDIATE;"),
        migrations.RunPython(add_missing_persons),
        migrations.RunSQL(sql="SET CONSTRAINTS ALL DEFERRED;"),
        migrations.AlterField(
            model_name="sportperson",
            name="profile",
            field=models.OneToOneField(
                on_delete=django.db.models.deletion.CASCADE, related_name="sport_profile", to="directory.person"
            ),
        ),
    ]
